#ifndef GUARD_audiobuffer_h
#define GUARD_audiobuffer_h
/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file audiobuffer.h
/// @brief Class for audio buffering header file.

#include <glibmm.h>
#include <sigc++/sigc++.h>

#include "fwd.h"
#include "effect.h"

/// @brief Namespace for audio class.
namespace Audio {

    /// @brief Class for audio buffering.
    template <class AudioPlayer>
    class AudioBuffer
    {
    public:
        AudioBuffer(const EffectorPreference& pref, AudioPlayer& audio);

        void Start(unsigned int bits, unsigned int channels, unsigned int sampleRate);

        void ConnectSignalTimerUpdate(const sigc::slot<void>& timerUpdateSlot);

        /// @brief Check pause state.
        /// @return If buffer play is paused, return true.
        bool GetPaused() const;

        /// @brief Get current play time.
        long int GetCurrentTime() const;

        /// @brief Toggle pause state.
        ///
        /// If you change player's state in pause, you should use
        /// skipPlayBufferedSample to true. Else you can hear buffered
        /// sample. 
        ///
        /// In short, you should use skipPlayBufferedSample
        /// if you not just unpause player.
        /// @param skipPlayBufferedSample Skip play buffered sample.
        void PauseToggle(bool skipPlayBufferedSample);

        void Unpause();

        void Push(float** const samples, unsigned int bytes);
        void Push(double* const samples, unsigned int bytes);
        void Push(mad_fixed_t* const left, unsigned int size);
        void Push(mad_fixed_t* const left, mad_fixed_t* const right, unsigned int size);
        void Push(short int* const samples, unsigned int size);

        void Close(bool flush);

        /// @brief Set current time to sync seek time.
        void SetCurrentTime(long int seconds);

    private:
        AudioBuffer(const AudioBuffer&);
        AudioBuffer operator=(const AudioBuffer&);

        // mad decoder output size is 4608, ogg vorbis output size is variable,
        // ffmpeg decoder output size is 4096. So i take 4608 * 31, and will 
        // work find.
        static const unsigned int internalBufferSize_ = 4608 * 31;
        static const unsigned int limitBufferSize_ = 4608 * 30;

        bool playLoopStart_;			///< For check play buffer is started. 
        bool playLoopStop_;			///< To stop play buffer thread.
        bool playLoopPause_;			///< Play buffer pause state. 
        bool skipBufferFlush_;			///< For skip play buffered samples.

        static const unsigned int outputBufferSize_ = 40960;
        unsigned int singleSampleSize_;		///< Size of single sample. Equal to bits / 8 * channels.
        unsigned int bufferSize_;		///< Size of buffer for effector and dithering.
        unsigned int bufferCount_;		///< Size of filled buffer.

        unsigned int sampleRate_;
        unsigned int channels_;

        Glib::Mutex playLoopMutex_;		///< Mutex to protect buffer.

        long int previousSeconds_;
        float currentSeconds_;
        float secondsPerOutputBuffer_;

        Glib::Cond playLoopCondPop_;		///< Condition variable for pop buffer.
        Glib::Cond playLoopCondPush_;		///< Condition variable for push buffer.

        Glib::Dispatcher signalTimerUpdate_;

        Glib::Thread* playLoopThread_;		///< Thread to play buffer.

        Effect::Processor processor_;

        const EffectorPreference& pref_;
        AudioPlayer& audio_;

        char outputBuffer_[outputBufferSize_];
        char buffer_[internalBufferSize_];			///< Buffer for play.

        void InternalPlayLoop();

        void InternalBufferPush(const char* buf);

        void ResetEffectorIfNeed();

        void StartPlayThread()
        {
            playLoopThread_ = Glib::Thread::create(sigc::mem_fun(*this, &AudioBuffer::InternalPlayLoop), true);

            playLoopStart_ = true;
        }

        /// @brief Reset member variables.
        void ResetTimer()
        {
            currentSeconds_ = 0.0f;
            previousSeconds_ = 0;
        }
    };

    template <class AudioPlayer>
    inline void AudioBuffer<AudioPlayer>::ConnectSignalTimerUpdate(const sigc::slot<void>& timerUpdateSlot)
    {
        signalTimerUpdate_.connect(timerUpdateSlot);
    }

    template <class AudioPlayer>
    inline bool AudioBuffer<AudioPlayer>::GetPaused() const
    {
        return playLoopPause_;
    }

    template <class AudioPlayer>
    inline long int AudioBuffer<AudioPlayer>::GetCurrentTime() const
    {
        return previousSeconds_;
    }

    /// @brief Set current time to sync seek time.
    template <class AudioPlayer>
    inline void AudioBuffer<AudioPlayer>::SetCurrentTime(long int seconds)
    {
        Glib::Mutex::Lock lock(playLoopMutex_);

        currentSeconds_ = seconds;
    }

    template <class AudioPlayer>
    inline void AudioBuffer<AudioPlayer>::Unpause()
    {
        if(GetPaused())
            PauseToggle(false);
    }
}

#endif
