#ifndef GUARD_effect_h
#define GUARD_effect_h
/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
//
// Mersenne Twister random number generator (MersenneTwister.h) is from
// 
// http://www-personal.engin.umich.edu/~wagnerr/MersenneTwister.html
//
// See also http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html
//
// for orig developer.
//
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file effect.h
/// @brief Effector functions header file.

#include <cmath>

#include <vector>

#include <mad.h>

#include "MersenneTwister.h"

class Preference;

/// @brief Namespace for effector class.
namespace Effect {

    struct Base {
        Base(): xn1(0.0), xn2(0.0), yn1(0.0), yn2(0.0) {}

        void Reset()
        {
            xn1 = xn2 = yn1 = yn2 = 0.0;
        }
	    
        double xn1, xn2, yn1, yn2;
    };

    /// @brief Class for stereo signal.
    struct Stereo {
        Stereo(const double& l, const double& r): left(l), right(r) {}
	    
        Stereo(): left(0.0), right(0.0) {}
	    
        double left, right;
    };

    template <class T>
    class BiquadCoeff {
    public:
        BiquadCoeff(): a0_(0.0), a1_(0.0), a2_(0.0), b1_(0.0), b2_(0.0), out_(0.0)
        {}
	
        ~BiquadCoeff() {}

        /// @brief Calculate coeffecients.
        /// @param sampleRate Sample rate.
        /// @param freq Frequency.
        /// @param gain Gain.
        /// @param Q Shape.
        void Calc(int sampleRate, int freq, float gain, double Q) {}

        /// @brief Apply to sample.
        void Apply(double& orig, Base& base) const
        {
            InternalProcess(orig, base);

            orig = out_;
        }

        /// @brief Return applied sample.
        double Return(double orig, Base& base) const
        {
            InternalProcess(orig, base);

            return out_;
        }

        /// @brief Reset member variable for static declared variable.
        void Reset()
        {
            a0_ = a1_ = a2_ = b1_ = b2_ = out_ = 0.0;
        }

    private:
        double a0_, a1_, a2_, b1_, b2_;
        mutable double out_;
        
        void InternalProcess(double orig, Base& base) const
        {
            out_ = a0_ * orig + a1_ * base.xn1 + a2_ * base.xn2
                - b1_ * base.yn1 - b2_ * base.yn2;
		
            base.xn2 = base.xn1;
            base.xn1 = orig;
            base.yn2 = base.yn1;
            base.yn1 = out_;
        }
    };

    enum LowShelf {};
    enum HighShelf {};
    enum Peak {};
    enum LowPass {};
    enum HighPass {};

    /// @brief Calculate coeffecients for low shelf EQ.
    template <>
    void BiquadCoeff<LowShelf>::Calc(int sampleRate, int freq, float gain, double Q);

    /// @brief Calculate coeffecients for high shelf EQ.
    template <>
    void BiquadCoeff<HighShelf>::Calc(int sampleRate, int freq, float gain, double Q);

    /// @brief Calculate coeffecients for peak EQ.
    template <>
    void BiquadCoeff<Peak>::Calc(int sampleRate, int freq, float gain, double Q);

    /// @brief Calculate coeffecients for low pass filter.
    template <>
    void BiquadCoeff<LowPass>::Calc(int sampleRate, int freq, float gain, double Q);

    /// @brief Calculate coeffecients for high pass filter.
    template <>
    void BiquadCoeff<HighPass>::Calc(int sampleRate, int freq, float gain, double Q);

    class Effector {
    public:
        virtual ~Effector() {}

        void Apply(double& sampleL, double& sampleR);
        void Calc(const Preference& pref, unsigned int sampleRate);
        void Reset();

    private:
        virtual void DoApply(double& sampleL, double& sampleR) = 0;
        virtual void DoCalc(const Preference& pref, unsigned int sampleRate) = 0;
        virtual void DoReset() = 0;
    };

    class Dummy : public Effector {
    private:
        virtual void DoApply(double& sampleL, double& sampleR);
        virtual void DoCalc(const Preference& pref, unsigned int sampleRate);
        virtual void DoReset();
    };

    class BassBoost : public Effector {
    private:
        BiquadCoeff<LowShelf> low100_;
        Base low100Left_, low100Right_;
        BiquadCoeff<Peak> peak40_;
        Base peak40Left_, peak40Right_;

        void DoApply(double& sampleL, double& sampleR);
        void DoCalc(const Preference& pref, unsigned int sampleRate);
        void DoReset();
    };

    class BassBoost2 : public Effector {
    private:
        BiquadCoeff<Peak> peak80_;
        Base peak80Left_, peak80Right_;
        BiquadCoeff<LowShelf> low199_;
        Base low199Left_, low199Right_;
        BiquadCoeff<HighShelf> high1309_;
        Base high1309Left_, high1309Right_;

        void DoApply(double& sampleL, double& sampleR);
        void DoCalc(const Preference& pref, unsigned int sampleRate);
        void DoReset();
    };

    class TrebleBoost : public Effector {
    private:
        BiquadCoeff<HighShelf> high5k_;
        Base high5kLeft_, high5kRight_;

        void DoApply(double& sampleL, double& sampleR);
        void DoCalc(const Preference& pref, unsigned int sampleRate);
        void DoReset();
    };

    class GraphicEQ : public Effector {
    private:
        BiquadCoeff<Peak> peak50_;
        Base peak50Left_, peak50Right_;

        BiquadCoeff<Peak> peak100_;
        Base peak100Left_, peak100Right_;

        BiquadCoeff<Peak> peak200_;
        Base peak200Left_, peak200Right_;

        BiquadCoeff<Peak> peak400_;
        Base peak400Left_, peak400Right_;

        BiquadCoeff<Peak> peak800_;
        Base peak800Left_, peak800Right_;

        BiquadCoeff<Peak> peak1500_;
        Base peak1500Left_, peak1500Right_;

        BiquadCoeff<Peak> peak3k_;
        Base peak3kLeft_, peak3kRight_;

        BiquadCoeff<Peak> peak5k_;
        Base peak5kLeft_, peak5kRight_;

        BiquadCoeff<Peak> peak7k_;
        Base peak7kLeft_, peak7kRight_;

        BiquadCoeff<Peak> peak10k_;
        Base peak10kLeft_, peak10kRight_;

        BiquadCoeff<Peak> peak15k_;
        Base peak15kLeft_, peak15kRight_;

        void DoApply(double& sampleL, double& sampleR);
        void DoCalc(const Preference& pref, unsigned int sampleRate);
        void DoReset();
    };

    class StereoEnhance : public Effector {
    private:
        double stereoEnhanceAmount_;

        void DoApply(double& sampleL, double& sampleR);
        void DoCalc(const Preference& pref, unsigned int sampleRate);
        void DoReset();
    };

    class Rhythme : public Effector {
    private:
        int prevCount_;
        double prevAvg_[2];
        double feedRhythme_;
        double rhythmeAmount_;

        void DoApply(double& sampleL, double& sampleR);
        void DoCalc(const Preference& pref, unsigned int sampleRate);
        void DoReset();
    };

    class Room : public Effector {
    private:
        int prevRoomCount_;
        double roomAmount_;
        double roomAmountSquare_;
        BiquadCoeff<LowPass> lowPass_;
        Base lowPassLeft_, lowPassRight_;
        double prevLeftSample_[5];
        double prevRightSample_[5];

        void DoApply(double& sampleL, double& sampleR);
        void DoCalc(const Preference& pref, unsigned int sampleRate);
        void DoReset();
    };

    class Room2 : public Effector {
    private:
        int room2Count_[2];
        double room2PrevSample_[1764];

        BiquadCoeff<LowShelf> low35_;
        Base low35Left_, low35Right_;

        BiquadCoeff<LowShelf> low62_;
        Base low62Left_, low62Right_;

        BiquadCoeff<LowShelf> low279_;
        Base low279Left_, low279Right_;

        BiquadCoeff<LowShelf> low852_;
        Base low852Left_, low852Right_;

        BiquadCoeff<HighShelf> high3713_;
        Base high3713Left_, high3713Right_;

        BiquadCoeff<HighShelf> high13545_;
        Base high13545Left_, high13545Right_;

        double room2CrossAmplifyMount_;

        void DoApply(double& sampleL, double& sampleR);
        void DoCalc(const Preference& pref, unsigned int sampleRate);
        void DoReset();
    };

    class Harmony : public Effector {
    private:
        int prevHarmonyCount_;
        double harmonyAmount_;
        double harmonyAmountSquare_;
        BiquadCoeff<HighPass> highPass_;
        Base highPassLeft_, highPassRight_;
        double prevLeftSample_[3];
        double prevRightSample_[3];

        void DoApply(double& sampleL, double& sampleR);
        void DoCalc(const Preference& pref, unsigned int sampleRate);
        void DoReset();
    };


    class Distortion : public Effector {
    private:
        double distortionAmount_;
        double distortionMutiple_;

        void DoApply(double& sampleL, double& sampleR);
        void DoCalc(const Preference& pref, unsigned int sampleRate);
        void DoReset();
    };

    class EffectorCreator {
    public:
        static const unsigned int BASSBOOST = 0;
        static const unsigned int BASSBOOST2 = 1;
        static const unsigned int TREBLEBOOST = 2;
        static const unsigned int GRAPHICEQ = 3;
        static const unsigned int STEREOENHANCE = 4;
        static const unsigned int RHYTHME = 5;
        static const unsigned int ROOM = 6;
        static const unsigned int ROOM2 = 7;
        static const unsigned int HARMONY = 8;
        static const unsigned int DISTORTION = 9;

        Effector* Create(unsigned int type)
        {
            switch(type) {
            case DISTORTION :
                return new Distortion;

            case HARMONY :
                return new Harmony;

            case ROOM2 :
                return new Room2;

            case ROOM :
                return new Room;

            case RHYTHME :
                return new Rhythme;

            case STEREOENHANCE :
                return new StereoEnhance;

            case GRAPHICEQ :
                return new GraphicEQ;

            case TREBLEBOOST :
                return new TrebleBoost;

            case BASSBOOST2 :
                return new BassBoost2;

            case BASSBOOST :
                return new BassBoost;

            default :
                return new Dummy;
            };
        }
    };

#ifdef WORDS_BIGENDIAN

    class ShortIntToChar {
    public:
        void operator()(short int in, char* out)
        {
            *out = in >> 8;
            *(out+1) = in & 0xff;
        }
    };

#else

    class ShortIntToChar {
    public:
        void operator()(short int in, char* out)
        {
            *out = in & 0xff;
            *(out+1) = in >> 8;
        }
    };

#endif

    struct Dither {

        Dither()
        {
            Reset();
        }

        unsigned int dPos_;
        double dBuffer_[9];
        double dTState_;

        void Reset();
    };

    /// @brief Simple modulo function.
    inline unsigned int Mod(unsigned int x, unsigned int n)
    {
        return (x > n) ? x - n : x;
    }

    class Dither16Bit {
    public:
        Dither16Bit()
            : ditherResult_(0.0)
        {}

        short int operator()(double& sample, Dither& dither);

        void Convert(double& sample, Dither& dither, short int* out);

    private:
        void Dithering(double& sample, Dither& dither)
        {
            double ditherRand = mtRand_();

            sample = sample * 32767.5
                - 0.4191 * dither.dBuffer_[Mod((dither.dPos_+8), 9)]
                + 1.632  * dither.dBuffer_[Mod((dither.dPos_+7), 9)]
                - 3.263  * dither.dBuffer_[Mod((dither.dPos_+6), 9)]
                + 5.032  * dither.dBuffer_[Mod((dither.dPos_+5), 9)]
                - 6.639  * dither.dBuffer_[Mod((dither.dPos_+4), 9)]
                + 7.184  * dither.dBuffer_[Mod((dither.dPos_+3), 9)]
                - 6.214  * dither.dBuffer_[Mod((dither.dPos_+2), 9)]
                + 4.684  * dither.dBuffer_[Mod((dither.dPos_+1), 9)]
                - 2.847  * dither.dBuffer_[(dither.dPos_)];

            ditherResult_ = rint(sample + ditherRand - dither.dTState_);

            dither.dPos_ = Mod((dither.dPos_+8), 9);
            dither.dBuffer_[dither.dPos_] = sample - ditherResult_;
            dither.dTState_ = ditherRand;
        }

        MTRand mtRand_;
        double ditherResult_;
    };

    class Clip {
    public:
        void Apply(double& sample)
        {
            if(sample > 1.0)
                UpLimit(sample);
	
            else if(sample < -1.0)
                DownLimit(sample);
        }

        void Apply(double& sampleL, double& sampleR)
        {
            Apply(sampleL);
            Apply(sampleR);
        }

        void Reset()
        {
            peakSample_ = 0.0;
            feedGain_ = 1.0;
        }

        double GetFeedGain()
        {
            return feedGain_;
        }

    private:
        double peakSample_;
        double feedGain_;

        void UpLimit(double& sample);
        void DownLimit(double& sample);
    };

    class Amplify {
    public:
        void Apply(double& sampleL, double& sampleR, float gain)
        {
            sampleL *= gain;
            sampleR *= gain;
        }

        void ApplyWithClip(double& sampleL, double& sampleR, float gain)
        {
            Apply(sampleL, sampleR, gain);

            clip_.Apply(sampleL, sampleR);
        }

        void ApplyInDecibel(double& sampleL, double& sampleR, float gain)
        {
            ApplyWithClip(sampleL, sampleR, pow(10.0, gain * 0.05));
        }

        void ApplyFeedGain(double& sampleL, double& sampleR)
        {
            ApplyWithClip(sampleL, sampleR, clip_.GetFeedGain());
        }

        void ApplyClip(double& sampleL, double& sampleR)
        {
            clip_.Apply(sampleL, sampleR);
        }

        void Reset()
        {
            clip_.Reset();
        }

    private:
        Clip clip_;
    };

    class Processor {
    public:
        Processor()
            : iterBegin_(effectors_.end()),
              iterEnd_(effectors_.end())
        {}

        ~Processor()
        {
            DeleteEffectors();
        }

        void EffectorStereo(float** const samples, unsigned int size, char* output);
        void EffectorStereo(double* const samples, unsigned int size, char* output);
        void EffectorStereo(mad_fixed_t* const left, mad_fixed_t* const right, unsigned int size, char* output);
        void EffectorStereo(short int* const left, unsigned int size, char* output);
        void DitherStereo(float** const samples, unsigned int size, char* output);
        void DitherStereo(double* const samples, unsigned int size, char* output);
        void DitherStereo(mad_fixed_t* const left, mad_fixed_t* const right, unsigned int size, char* output);
        void DitherStereo(short int* const left, unsigned int size, char* output);

        void EffectorMono(float** const samples, unsigned int size, char* output);
        void EffectorMono(double* const samples, unsigned int size, char* output);
        void EffectorMono(mad_fixed_t* const samples, unsigned int size, char* output);
        void EffectorMono(short int* const samples, unsigned int size, char* output);
        void DitherMono(float** const samples, unsigned int size, char* output);
        void DitherMono(double* const samples, unsigned int size, char* output);
        void DitherMono(mad_fixed_t* const samples, unsigned int size, char* output);
        void DitherMono(short int* const samples, unsigned int size, char* output);

        void Update(const Preference& pref, unsigned int sampleRate);
        void Calc(const Preference& pref, unsigned int sampleRate);
        void Reset();

    private:
        std::vector<Effector*> effectors_;
        std::vector<Effector*>::iterator iterBegin_;
        std::vector<Effector*>::iterator iterEnd_;
        EffectorCreator effectorCreator_;

        Amplify amplify_;

        ShortIntToChar shortIntToChar_;
        Dither16Bit dither_;
        Dither ditherLeft_, ditherRight_;

        void ApplyDither(double& sample, Dither& dither, char*& out)
        {
            dither_.Convert(sample, dither, reinterpret_cast<short int*>(out));
            out += 2;
        }

        void ApplyDither(double& left, double& right, char*& out)
        {
            ApplyDither(left, ditherLeft_, out);
            ApplyDither(right, ditherRight_, out);
        }

        void DeleteEffectors();
        void RemoveAll();

        void PutEffector(const Preference& pref, unsigned int sampleRate, unsigned int type);

        void ApplyEffector(double& left, double& right);
    };

    inline void Effector::Apply(double& sampleL, double& sampleR)
    {
        DoApply(sampleL, sampleR);
    }

    inline void Effector::Calc(const Preference& pref, unsigned int sampleRate)
    {
        DoCalc(pref, sampleRate);
    }

    inline void Effector::Reset()
    {
        DoReset();
    }
}

#endif
