#ifndef GUARD_fixedallocator_h
#define GUARD_fixedallocator_h
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// you know, it from 'Modern C++ Design' by Andrei.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file fixedallocator.h
/// @brief Class for fixed size allocator header file.
///
/// Fixed size allocator from 'Modern C++ Design' by Andrei.

#include <cstddef>
#include <vector>

#define DEFAULT_CHUNK_SIZE 4096
#define MAX_SMALL_OBJECT_SIZE 64

class FixedAllocator
{
    public:
	explicit FixedAllocator(std::size_t blockSize = 0);
	FixedAllocator(const FixedAllocator&);
	FixedAllocator& operator=(const FixedAllocator&);
	~FixedAllocator();

	void Swap(FixedAllocator& rhs);

	void* Allocate();
	void Deallocate(void* p);

	std::size_t BlockSize() const
	{ return blockSize_; }

	bool operator<(std::size_t rhs) const
	{ return BlockSize() < rhs; }

    private:
	struct Chunk
	{
	    void Init(std::size_t blockSize, unsigned char blocks);
	    void* Allocate(std::size_t blockSize);
	    void Deallocate(void* p, std::size_t blockSize);
	    void Reset(std::size_t blockSize, unsigned char blocks);
	    void Release();
	    unsigned char* pData_;
	    unsigned char firstAvailableBlock_;
	    unsigned char blocksAvailable_;

	};

	void DoDeallocate(void* p);
	Chunk* VicinityFind(void* p);

	std::size_t blockSize_;
	unsigned char numBlocks_;

	typedef std::vector<Chunk> Chunks;
	Chunks chunks_;
	Chunk* allocChunk_;
	Chunk* deallocChunk_;

	mutable const FixedAllocator* prev_;
	mutable const FixedAllocator* next_;

};

class SmallObjAllocator
{
    public:
	SmallObjAllocator(
		std::size_t chunkSize, 
		std::size_t maxObjectSize);

	void* Allocate(std::size_t numBytes);
	void Deallocate(void* p, std::size_t size);

    private:
	SmallObjAllocator(const SmallObjAllocator&);
	SmallObjAllocator& operator=(const SmallObjAllocator&);

	typedef std::vector<FixedAllocator> Pool;
	Pool pool_;
	FixedAllocator* pLastAlloc_;
	FixedAllocator* pLastDealloc_;
	std::size_t chunkSize_;
	std::size_t maxObjectSize_;
};

struct MySmallObjAllocator : public SmallObjAllocator
{
    MySmallObjAllocator()
	: SmallObjAllocator(DEFAULT_CHUNK_SIZE, MAX_SMALL_OBJECT_SIZE) {}	
};	

MySmallObjAllocator& SmallObjAllocatorInstance();

class SmallObject
{
    public:
	virtual ~SmallObject() {} 

	static void* operator new(std::size_t size)
	{
	    return SmallObjAllocatorInstance().Allocate(size);
	}

	static void operator delete(void* p, std::size_t size)
	{
	    return SmallObjAllocatorInstance().Deallocate(p, size);
	}

};

#endif
