#ifndef GUARD_itemselector_h
#define GUARD_itemselector_h
/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file itemselector.h
/// @brief Class to get item from playlist header file.

#include <string>

#include "enums.h"
#include "playlist.h"

/// @brief Class to get item from playlist.
class ItemSelector
{
public:
    ItemSelector(GetItem& getItem)
        : listState_(Sagua::LIST_EMPTY),
          getItem_(getItem)
    {}

    virtual ~ItemSelector();

    /// @brief Get next item.
    /// @return Success of getting item.
    virtual bool GetNext(std::string& item) = 0;
    /// @brief Get previous item.
    /// @return Success of getting item.
    virtual bool GetPrev(std::string& item) = 0;

protected:
    Sagua::ListState listState_;
    GetItem& getItem_;
};

class RepeatSelector : public ItemSelector
{
public:
    RepeatSelector(GetItem& getItem)
        : ItemSelector(getItem)
    {}

    virtual bool GetNext(std::string& item);
    virtual bool GetPrev(std::string& item);

protected:
    /// @brief Get next item circularly.
    bool GetCircularNext(std::string& item)
    {
        item = getItem_.GetNext(listState_);

        if(Sagua::LIST_EMPTY == listState_)
            return false;

        if(Sagua::LIST_LAST == listState_)
            item = getItem_.GetFirst(listState_);

        return true;
    }

    /// @brief Get previous item circularly.
    bool GetCircularPrev(std::string& item)
    {
        item = getItem_.GetPrev(listState_);

        if(Sagua::LIST_EMPTY == listState_)
            return false;

        if(Sagua::LIST_FIRST == listState_)
            item = getItem_.GetLast(listState_);

        return true;
    }
};

class NormalSelector : public ItemSelector
{
public:
    NormalSelector(GetItem& getItem)
        : ItemSelector(getItem)
    {}

    virtual bool GetNext(std::string& item);
    virtual bool GetPrev(std::string& item);

private:
    bool GetOnewayNext(std::string& item)
    {
        item = getItem_.GetNext(listState_);

        return Sagua::LIST_LAST != listState_ && Sagua::LIST_EMPTY != listState_;
    }

    bool GetOnewayPrev(std::string& item)
    {
        item = getItem_.GetPrev(listState_);

        return Sagua::LIST_FIRST != listState_ && Sagua::LIST_EMPTY != listState_;
    }
};

class RandomSelector : public ItemSelector
{
public:
    RandomSelector(GetItem& getItem)
        : ItemSelector(getItem)
    {}

    virtual bool GetNext(std::string& item);
    virtual bool GetPrev(std::string& item);

private:
    bool GetRandom(std::string& item)
    {
        item = getItem_.GetRandom(listState_);

        return Sagua::LIST_EMPTY != listState_;
    }
};

/// @brief Class for default selector, always fail selector.
class EmptySelector : public ItemSelector
{
public:
    EmptySelector(GetItem& getItem)
        : ItemSelector(getItem)
    {}

    virtual bool GetNext(std::string& item);
    virtual bool GetPrev(std::string& item);
};

/// @brief Class to create item selector.
class ItemSelectorCreator
{
public:
    static const unsigned int NORMAL = 0;
    static const unsigned int REPEAT = 1;
    static const unsigned int RANDOM = 2;
    static const unsigned int EMPTY = 3;

    std::auto_ptr<ItemSelector> Create(unsigned int type, GetItem& getItem);
};

#endif
