#ifndef GUARD_mainwin_h
#define GUARD_mainwin_h
/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file mainwin.h
/// @brief Class for main window header file.

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <vector>
#include <string>

#include <glibmm.h>
#include <gdkmm.h>
#include <gtkmm/window.h>
#include <gtkmm/box.h>
#include <gtkmm/label.h>
#include <gtkmm/menu.h>
#include <gtkmm/menubar.h>
#include <gtkmm/checkmenuitem.h>
#include <gtkmm/alignment.h>
#include <gtkmm/scale.h>
#include <gtkmm/button.h>
#include <gtkmm/toolbar.h>
#include <gtkmm/toolbutton.h>
#include <gtkmm/eventbox.h>
#include <gtkmm/statusbar.h>
#include <gtkmm/arrow.h>

#include "fwd.h"

#include "imagebutton.h"
#include "misc.h"
#include "enums.h"

class MainWnd
{
public:
    virtual ~MainWnd() {}

    virtual Gtk::Window& GetWidget() = 0;

    virtual void TimeUpdate() = 0;

    virtual void SaveState() = 0;

    virtual void SyncMixerWidget() = 0;

    virtual void ChangeCompactBgPixmap() = 0;

    virtual void SetStatusbar(const Glib::ustring& text) = 0;

    virtual void SetStatusbarText(const Glib::ustring& text) = 0;

    virtual void UpdateStatusbar() = 0;

    virtual void UpdateStatusbarTimeoutClear() = 0;

    virtual void ClearStatusbar() = 0;

    virtual void TimeoutClearStatusbar() = 0;

    virtual void ShowAudioOpenFailDlg() = 0;

    virtual void SetPlayMode() = 0;

    virtual void SetRepeatCurrent() = 0;

    virtual void UpdateItemRead() = 0;

    virtual void UpdatePlay() = 0;

    virtual void UpdateStop() = 0;

    virtual void UpdateItemRemoved() = 0;
};

/// @brief Class for main window.
/// @todo Consider use Gtk::Stock::MEDIA_* instead custom button image when
/// over Gtkmm 2.6.x
class MainWndImpl : public Gtk::Window, public MainWnd
{
public:
    MainWndImpl(MainWndInterface& mainWndInterface, SettingInterface& setting, PlaylistDisplay& playlist, AddItem& addItem, DirPlayWnd& dirPlayWnd);

    virtual Gtk::Window& GetWidget();

    virtual void TimeUpdate();

    virtual void SaveState();

    /// @brief Set to skip time display update.
    void SetSkipTimeUpdate(bool active)
    {
        skipTimeUpdate_ = active;
    }

    virtual void SyncMixerWidget();

    /// @brief Change background pixmap used when compact mode.
    virtual void ChangeCompactBgPixmap();

    /// @brief Set statusbar text and update it.
    virtual void SetStatusbar(const Glib::ustring& text);

    /// @brief Set statusbar text.
    ///
    /// For inter thread dispatcher.
    virtual void SetStatusbarText(const Glib::ustring& text);

    /// @brief Update statusbar to stored text.
    ///
    /// For inter thread dispatcher.
    /// @sa SetStatusbarText
    virtual void UpdateStatusbar();

    /// @brief Update statusbar to stored text and clear at timeout.
    virtual void UpdateStatusbarTimeoutClear();

    /// @brief Clear statusbar.
    virtual void ClearStatusbar();

    /// @brief Clear statusbar at timeout.
    virtual void TimeoutClearStatusbar();

    virtual void ShowAudioOpenFailDlg();

    virtual void SetPlayMode();
    virtual void SetRepeatCurrent();

    virtual void UpdateItemRead();

    virtual void UpdatePlay();

    virtual void UpdateStop();

    virtual void UpdateItemRemoved();

private:
    Gtk::VBox mainBox_, playerBox_, compactPlayCtlBox_, arrowBoxWithVolumn_;
    Gtk::HBox playCtlBox_, compactPlayCtlHBox_, timeDisplayBox_, arrowBox_, playerBoxWithVolumn_, viewBox_;
    Gtk::MenuBar menuBar_;
    Gtk::Menu fileMenu_, editMenu_, viewMenu_, helpMenu_;
    Gtk::CheckMenuItem repeatPlayMenu_, randomPlayMenu_, repeatCurrentPlayMenu_, listDisplayMenu_, browserDisplayMenu_, statusbarDisplayMenu_;
    Gtk::Label displayLabel_, timeLabel_, totalTimeLabel_;
    Gtk::Alignment titleDisplayAlignment_, seekScaleAlignment_, volumnAlignment_;
    Gtk::EventBox titleDisplayEventBox_, leftArrowEventBox_, rightArrowEventBox_, upArrowEventBox_, downArrowEventBox_;
    Gtk::HScale seekScale_;
    Gtk::VScale volumnScale_;
    ImageButton playImgBtn_, pauseImgBtn_, stopImgBtn_, prevImgBtn_, nextImgBtn_;
    ImageButton compactPlayImgBtn_, compactPauseImgBtn_, compactStopImgBtn_, compactPrevImgBtn_, compactNextImgBtn_;

    Gtk::Arrow leftArrow_, rightArrow_, upArrow_, downArrow_;

    Gtk::Statusbar statusbar_;
    Glib::ustring statusbarText_;

    const int defaultWidth_, defaultHeight_;
    const int defaultCompactWidth_, defaultCompactHeight_;
    const int defaultFoldHeight_;
    const int defaultTitleWidth_;
    int prevWidth_, prevHeight_;
    int maxScreenWidth_, maxScreenHeight_;

    bool compactMode_;

    bool skipTimeUpdate_;

    Glib::RefPtr<Gtk::RcStyle> mainWndDefaultStyle_;
    Glib::RefPtr<Gtk::RcStyle> titleDisplayEventBoxDefaultStyle_;

    PlaylistDisplay& playlist_;
    DirPlayWnd& dirPlayWnd_;

    sigc::connection* clearStatusbarTimeout_;

    MainWndInterface& mainWndInterface_;
    SettingInterface& setting_;

    virtual bool on_delete_event(GdkEventAny* event);
    virtual bool on_map_event(GdkEventAny* event);
    virtual bool on_unmap_event(GdkEventAny* event);

    virtual bool on_configure_event(GdkEventConfigure* ev);

    void OnMenuViewStatusbar();
    void OnMenuQuit();

    void ToggleCompactView();

    void OnScaleSeek(const double&);

    bool OnButtonPressInDisplayLabel(GdkEventButton* ev);

    bool OnLeftArrowPress(GdkEventButton* ev);
    bool OnRightArrowPress(GdkEventButton* ev);
    bool OnUpArrowPress(GdkEventButton* ev);
    bool OnDownArrowPress(GdkEventButton* ev);

    void UpdateDisplay();

    void SetBackgroundPixmapInCompactMode();
    void UnsetBackgroundPixmapInCompactMode();

    void ClearDisplayTitle();
    void DisplayTitle(const Glib::ustring& title);

    void Packing(AddItem& addItem);
    void InitialDisplay();
    void ConnectSignals();

    /// @brief Display total time part.
    void DisplayTotalTime(const Glib::ustring& time)
    {
        totalTimeLabel_.set_label(" / " + time);
    }

    /// @brief Display time and set sensitive to seek bar.
    void DisplayVisualPanel()
    {
        timeLabel_.show();
        totalTimeLabel_.show();
        seekScale_.set_sensitive(true);
    }

    void ClearVisualPanel()
    {
        timeLabel_.hide();
        totalTimeLabel_.hide();
        seekScale_.set_sensitive(false);
    }

    /// @brief Set maximum range of seek bar.
    void SetSeekScaleRange(unsigned int max)
    {
        seekScale_.set_range(0.0, max);
    }

    void SetMouseCursorToWatch()
    {
        get_window()->set_cursor(Gdk::Cursor(Gdk::WATCH));
        while(Glib::MainContext::get_default()->iteration(false));
    }

    void RestoreMouseCursor()
    {
        get_window()->set_cursor();
    }

    Glib::ustring BoldWeightTitle(const Glib::ustring& title, bool normalSize)
    {
        return normalSize ?
            "<span weight=\"bold\">" + Glib::Markup::escape_text(title) + "</span>" :
            "<span weight=\"bold\" size=\"large\">" + Glib::Markup::escape_text(title) + "</span>";
    }


    void SetDisplayLabel(const Glib::ustring& string)
    {
        displayLabel_.set_label(string);
        displayLabel_.set_use_markup(true);
        displayLabel_.set_alignment(0.5, 0.5);
    }

    /// @brief Disconnect timeout handler and delete it.
    void DisconnectTimeoutClearStatusbar()
    {
        if(0 == clearStatusbarTimeout_)
            return;

        clearStatusbarTimeout_->disconnect();
        delete clearStatusbarTimeout_;
        clearStatusbarTimeout_ = 0;
    }

    /// @brief Clear statusbar function for timeout handler.
    bool TimeoutClearStatusbarSlot()
    {
        ClearStatusbar();

        delete clearStatusbarTimeout_;
        clearStatusbarTimeout_ = 0;

        return false;
    }

    /// @brief Get size from a physical screen in pixel.
    void GetScreenSize()
    {
        Glib::RefPtr<Gdk::Screen> screen(get_screen());
        maxScreenWidth_ = screen->get_width();
        maxScreenHeight_ = screen->get_height();
    }
};

#endif
