#ifndef GUARD_playlist_h
#define GUARD_playlist_h
/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file playlist.h
/// @brief Class for Playlist header file.

#include <string>
#include <cstring>
#include <queue>
#include <vector>
#include <map>

#include <gtkmm/treeview.h>
#include <gtkmm/treemodel.h>
#include <gtkmm/liststore.h>
#include <gtkmm/menu.h>

#include "fwd.h"
#include "misc.h"
#include "enums.h"

class GetItem
{
public:
    virtual ~GetItem() {}

    std::string GetCurrent(Sagua::ListState& listState);
    std::string GetNext(Sagua::ListState& listState);
    std::string GetPrev(Sagua::ListState& listState);
    std::string GetFirst(Sagua::ListState& listState);
    std::string GetLast(Sagua::ListState& listState);
    std::string GetRandom(Sagua::ListState& listState);

private:
    virtual std::string DoGetCurrent(Sagua::ListState& listState) = 0;
    virtual std::string DoGetNext(Sagua::ListState& listState) = 0;
    virtual std::string DoGetPrev(Sagua::ListState& listState) = 0;
    virtual std::string DoGetFirst(Sagua::ListState& listState) = 0;
    virtual std::string DoGetLast(Sagua::ListState& listState) = 0;
    virtual std::string DoGetRandom(Sagua::ListState& listState) = 0;
};

inline std::string GetItem::GetCurrent(Sagua::ListState& listState)
{
    return DoGetCurrent(listState);
}

inline std::string GetItem::GetNext(Sagua::ListState& listState)
{
    return DoGetNext(listState);
}

inline std::string GetItem::GetPrev(Sagua::ListState& listState)
{
    return DoGetPrev(listState);
}

inline std::string GetItem::GetFirst(Sagua::ListState& listState)
{
    return DoGetFirst(listState);
}

inline std::string GetItem::GetLast(Sagua::ListState& listState)
{
    return DoGetLast(listState);
}

inline std::string GetItem::GetRandom(Sagua::ListState& listState)
{
    return DoGetRandom(listState);
}

class PlaylistDisplay
{
public:
    virtual ~PlaylistDisplay() {}

    virtual void SetTitleWidth(unsigned int titleWidth) = 0;

    virtual Glib::ustring CurrentItemTitle() = 0;

    virtual Glib::ustring CurrentItemTime() = 0;

    virtual unsigned int CurrentItemTimeSeconds() = 0;
};

class Playlist : public GetItem, public PlaylistDisplay
{
public:
    virtual ~Playlist() {}

    virtual Gtk::Widget& GetWidget() = 0;

    virtual void AddFiles(const std::vector<std::string>& files, const bool& addListFile) = 0;

    virtual void AddDirectory(const std::string& directoryPath) = 0;
    virtual void AddDirectoryRecursive(const std::string& path) = 0;

#ifdef ENABLE_HTTP

    virtual void AddUrl(const std::string& destUrl) = 0;

#endif

    virtual void RemoveAll() = 0;

    virtual void ShowCurrentFileInfoDlg() const = 0;

    virtual void ForceUpdateToppage() = 0;
    virtual void ForceUpdatePage() = 0;

    sigc::signal0<bool>& SignalConvertUrlEncoding()
    {
        return signalConvertUrlEncoding_;
    }

protected:
    sigc::signal0<bool> signalConvertUrlEncoding_;
};

/// @brief Class for Playlist.
/// @todo Reduce relationship to MainWnd.
class PlaylistImpl : public Gtk::TreeView, public Playlist
{
    friend class ListFile;

public:
    PlaylistImpl(AddItem& addItem, PlayListControl& playCtl, unsigned int mainWndHeight = 300, unsigned int titleWidth = 470, const bool& loadlist = true);
    ~PlaylistImpl();

    virtual Gtk::Widget& GetWidget();

    virtual void AddFiles(const std::vector<std::string>& files, const bool& addListFile);

    virtual void AddDirectory(const std::string& directoryPath);
    virtual void AddDirectoryRecursive(const std::string& path);

#ifdef ENABLE_HTTP

    virtual void AddUrl(const std::string& destUrl);

#endif

    /// @brief Remove all items.
    virtual void RemoveAll();

    virtual void ForceUpdateToppage();
    virtual void ForceUpdatePage();

    virtual Glib::ustring CurrentItemTitle();

    virtual Glib::ustring CurrentItemTime();

    virtual unsigned int CurrentItemTimeSeconds();

    virtual void ShowCurrentFileInfoDlg() const;

    virtual void SetTitleWidth(unsigned int titleWidth);

private:
    class ModelColumns : public Gtk::TreeModel::ColumnRecord
    {
    public:
        ModelColumns()
        {
            add(tagParsed_);
            add(id_);
            add(filename_);
            add(title_);
            add(time_);
            add(timeSeconds_);
        }

        Gtk::TreeModelColumn<bool> tagParsed_;
        Gtk::TreeModelColumn<unsigned int> id_;
        Gtk::TreeModelColumn<std::string> filename_;
        Gtk::TreeModelColumn<Glib::ustring> title_;
        Gtk::TreeModelColumn<Glib::ustring> time_;
        Gtk::TreeModelColumn<unsigned int> timeSeconds_;
    };

    ModelColumns columns_;

    Gtk::Menu menuPopup_;
    Gtk::TreeModel::iterator pos_;

    Glib::Mutex tagParsingMutex_;
    Glib::Dispatcher signalRowToFill_;

    class EachRow;
    std::queue<EachRow*> rowQueue_;

    static const unsigned int mainWndPanelHeight_;
    static const unsigned int eachRowHeight_;
    unsigned int count_;				///< Number of added items.
    unsigned int tagProcessCount_;			///< Number of items reading tag at one thread.

    bool inTagParsing_;				///< Tag parsing thread is running.
    bool stopTagParsing_;				///< To stop tag parsing.

    Glib::RefPtr<Gtk::ListStore> treeModel_;

    const std::auto_ptr<TagFactory> tagFactory_;
    const std::auto_ptr<ListFactory> listFactory_;

    PlayListControl& playCtl_;

    Glib::Thread* tagParsingThread_;

    std::string DoGetCurrent(Sagua::ListState& listState);
    std::string DoGetNext(Sagua::ListState& listState);
    std::string DoGetPrev(Sagua::ListState& listState);
    std::string DoGetFirst(Sagua::ListState& listState);
    std::string DoGetLast(Sagua::ListState& listState);
    std::string DoGetRandom(Sagua::ListState& listState);

    virtual bool on_button_press_event(GdkEventButton* ev);
    virtual void on_row_activated(const Gtk::TreeModel::Path& path, Gtk::TreeViewColumn* column);
    virtual void on_set_scroll_adjustments(Gtk::Adjustment* hadj, Gtk::Adjustment* vadj);
    virtual void on_size_allocate(Gtk::Allocation& allocation);

    virtual void on_drag_data_received(const Glib::RefPtr<Gdk::DragContext>& context, int x, int y, const Gtk::SelectionData& selectionData, guint info, guint time);

    void AddFile(const std::string& filePath, const bool& addListFile);
    void AddFile(const std::string& filePath)
    {
        Gtk::TreeModel::Row row( *(treeModel_->append()) );
        row[columns_.tagParsed_] = false;
        row[columns_.id_] = ++count_;
        row[columns_.filename_] = filePath;
        row[columns_.timeSeconds_] = 1;
    }

    void AddFile(const std::string& filePath,
                 const Glib::ustring& title,
                 const Glib::ustring& time,
                 const unsigned int& seconds)
    {
        Gtk::TreeModel::Row row( *(treeModel_->append()) );
        row[columns_.tagParsed_] = true;
        row[columns_.id_] = ++count_;
        row[columns_.filename_] = filePath;
        // need convert to utf-8?
        row[columns_.title_] = filePath;
        // convert from time seconds ?
        row[columns_.time_] = time;
        row[columns_.timeSeconds_] = seconds;
    }


    void FillRowFromQueue(); 

    void GetInfoFromTag(Gtk::TreeModel::iterator& iter);

    void ShowFileInfoDlg(const std::string& filename) const;


    void OnMenuRemove();
    void OnMenuRemoveAll();
    void OnMenuSortFilename();
    void OnMenuInfo();

    void OnVScrolled();

    void LoadLastList();

#ifdef ENABLE_HTTP

    void GetListFromUrl(const std::string& destUrl,
                        const std::string& listFile);

#endif

    void LoadList(const std::string& filename);
    void SaveList();

    void StopThread();

    void CreatePopupMenuList(AddItem& addItem);
    void RegisterTagFactory();

    void RegisterListType();
    bool IsListType(const std::string& filePath);

    void AddDragedFiles(const std::string& dragedString);

    std::string EmptyNullFilename(Sagua::ListState& listState)
    {
        listState = Sagua::LIST_EMPTY;

        return "";
    }

    std::string FirstNullFilename(Sagua::ListState& listState)
    {
        listState = Sagua::LIST_FIRST;

        return "";
    }

    std::string LastNullFilename(Sagua::ListState& listState)
    {
        listState = Sagua::LIST_LAST;

        return "";
    }

    /// @brief Filename compare function to sort list.
    int FilenameCompare(const Gtk::TreeModel::iterator& a, const Gtk::TreeModel::iterator& b);


    /// @brief If list is empty, return true.
    bool IsListEmpty()
    {
        if(count_ == 0 || !treeModel_->children().begin())
            return true;

        return false;
    }

    /// @brief If stored iterator is last row, return true.
    bool IsPositionLastRow() const
    {
        return pos_ == --(treeModel_->children().end());
    }

    /// @brief If storef iterator is first row, return true.
    bool IsPositionFirstRow() const
    {
        return pos_ == treeModel_->children().begin();
    }

    /// @brief Calculate row number per page.
    void CalcTagProcessCount(const unsigned int& height)
    {
        tagProcessCount_ = height / eachRowHeight_ + 2;
    }

    /// @brief Fill row.
    void FillRow(Gtk::TreeModel::iterator& iter, const Glib::ustring& title, const Glib::ustring& totalTime, const unsigned long& timeSeconds)
    {
        (*iter)[columns_.title_] = title;
        (*iter)[columns_.time_] = totalTime;
        (*iter)[columns_.timeSeconds_] = timeSeconds;

        if(timeSeconds != 1)
            (*iter)[columns_.tagParsed_] = true;
    }


    /// @brief Set title column properties.
    void SetTitleColumnProperties(const unsigned int& titleWidth)
    {
        get_column(1)->set_resizable(true);
        get_column(1)->set_fixed_width(titleWidth);
        get_column(1)->set_sizing(Gtk::TREE_VIEW_COLUMN_FIXED);
    }

    Gtk::TreeModel::iterator GetFirstIter() const
    {
        return treeModel_->children().begin();
    }

    void GetRawInformation(const Gtk::TreeModel::iterator& iter, unsigned int& seconds, Glib::ustring& title, std::string& filename) const
    {
        seconds = (*iter)[columns_.timeSeconds_];
        title = (*iter)[columns_.title_];
        filename = (*iter)[columns_.filename_];
    }

    void SetFontSize(unsigned int size);

    void TagParsing(Gtk::TreeModel::iterator& iter);
    std::string Get(Gtk::TreeModel::iterator& iter);
    std::string Get(Gtk::TreeModel::iterator& iter, Sagua::ListState& listState);
    void MoveNthPosition(Gtk::TreeModel::iterator& iter, const unsigned int& number);
    unsigned int GetRandomNumber(Gtk::TreeModel::iterator& iter);

    Gtk::TreeModel::iterator GetTopPositionIterator();

    void MoveCursorAndScroll(const Gtk::TreeModel::iterator& iter);

    void ClearRowQueue();
};

class ListFile
{
public:
    virtual void Read(const std::string& filename, PlaylistImpl& output) = 0;
    virtual void Write(const PlaylistImpl& input, const std::string& filename) = 0;

protected:
    void AddFile(const std::string& filePath, PlaylistImpl& playlist)
    {
        playlist.AddFile(filePath, false);
    }

#ifdef ENABLE_HTTP

    void AddUrl(const std::string& filePath, PlaylistImpl& playlist)
    {
        playlist.AddUrl(filePath);
    }

#endif

    Gtk::TreeModel::iterator GetFirstIter(const PlaylistImpl& playlist) const
    {
        return playlist.GetFirstIter();
    }

    void GetRawInformation(const PlaylistImpl& playlist, const Gtk::TreeModel::iterator& iter, unsigned int& seconds, Glib::ustring& title, std::string& filename) const
    {
        playlist.GetRawInformation(iter, seconds, title, filename);
    }
};

class M3UList : public ListFile
{
public:
    virtual void Read(const std::string& filename, PlaylistImpl& output);
    virtual void Write(const PlaylistImpl& input, const std::string& filename);
};

class PLSList : public ListFile
{
public:
    virtual void Read(const std::string& filename, PlaylistImpl& output);
    virtual void Write(const PlaylistImpl& input, const std::string& filename);
};

std::auto_ptr<ListFile> CreateM3UList();
std::auto_ptr<ListFile> CreatePLSList();

class ListFactory
{
public:
    /// @brief Register id to callback map.
    /// @param id Identification string 
    /// @param creator callback to produce object
    /// @return Success of register.
    bool Register(const char* const id, std::auto_ptr<ListFile> (*creator)())
    {
        return callbacks_.insert( callback::value_type(id, creator) ).second;
    }

    /// @brief Unregister id from callback map.
    /// @param id Identification string.
    /// @return Success of unregister.
    bool Unregister(const std::string& id)
    {
        return callbacks_.erase(id) == 1;
    }

    /// @brief Check factory has filename's type.
    /// @param filename Filename to check.
    /// @return Return true if factory has filename's type.
    bool Exist(const std::string& filename) const
    {
        callback::const_iterator i = callbacks_.find(GetExt(filename));

        if(i != callbacks_.end())
            return true;
        else
            return false;	
    }

    /// @brief Create ListFile object.
    /// @param filename Filename to play.
    /// @return ListFile object.
    std::auto_ptr<ListFile> Create(const std::string& filename)
    {
        callback::const_iterator i = callbacks_.find(GetExt(filename));

        if(i != callbacks_.end())
            return (i->second)();
        else
            return std::auto_ptr<ListFile>();	
    }

    bool Empty() const
    {
        return callbacks_.empty();
    }

private:
    typedef std::map<std::string, std::auto_ptr<ListFile> (*)()> callback;
    callback callbacks_;	    ///< callback map
};

#endif
