#ifndef GUARD_preference_h
#define GUARD_preference_h
/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file preference.h
/// @brief Class for user preference setting header file.

#include <iosfwd>

#include <map>
#include <string>

class Preference
{
public:
    virtual ~Preference() {}

    const double& GetDouble(const char* key) const;

    float GetFloat(const char* key) const;

    bool GetBool(const char* key) const;

private:
    virtual const double& DoGetDouble(const char* key) const = 0;

    virtual float DoGetFloat(const char* key) const = 0;

    virtual bool DoGetBool(const char* key) const = 0;
};

class EffectorPreference : public Preference
{
public:
    EffectorPreference()
        : efxChanged_(true)
    {}

    virtual ~EffectorPreference() {}

    const std::string& GetString(const char* key) const;

    void Set(const char* key, bool value);

    void UpdateEffectUse()
    {
        if(GetBool("EfxRoom") ||
           GetBool("EfxRoom2") ||
           GetBool("EfxHarmony") ||
           GetBool("EfxRhythme") ||
           GetBool("EfxBassBoost") ||
           GetBool("EfxGraphiceq") ||
           GetBool("EfxBassBoost2") ||
           GetBool("EfxDistortion") ||
           GetBool("EfxTrebleBoost") ||
           GetBool("EfxStereoEnhance")) {

            effectUse_ = true;

            Set("Effect", true);

        } else {

            effectUse_ = false;

            Set("Effect", false);
        }
    }

    bool GetEffectUse() const
    {
        return effectUse_;
    }

    /// @brief Need to calculate some effector coefficient.
    bool GetEfxChanged() const
    {
        return efxChanged_;
    }

    /// @brief Set efxChanged to true.
    void EfxChangedSet() const
    {
        efxChanged_ = true;
    }
    /// @brief Set efxChanged to false.
    void EfxChangedUnset() const
    {
        efxChanged_ = false;
    }

protected:
    bool effectUse_;
    mutable bool efxChanged_;

private:
    virtual const std::string& DoGetString(const char* key) const = 0;

    virtual void DoSet(const char* key, bool value) = 0;
};

class SaguaPreference : public EffectorPreference
{
public:
    SaguaPreference();

    void Save(std::ostream& os);

    void Get(const char* key, std::string& dest) const
    {
        std::map<std::string, std::string>::const_iterator iter(stringElem_.find(key));

        if(iter != stringElem_.end())
            dest = iter->second;
    }

    void Get(const char* key, double& dest) const
    {
        std::map<std::string, double>::const_iterator iter(doubleElem_.find(key));

        if(iter != doubleElem_.end())
            dest = iter->second;
    }

    void Get(const char* key, float& dest) const
    {
        std::map<std::string, float>::const_iterator iter(floatElem_.find(key));

        if(iter != floatElem_.end())
            dest = iter->second;
    }

    void Get(const char* key, bool& dest) const
    {
        std::map<std::string, bool>::const_iterator iter(boolElem_.find(key));

        if(iter != boolElem_.end())
            dest = iter->second;
    }

    void GetAudioDriverOption(std::map<std::string, std::string>& dest);

    using EffectorPreference::Set;

    void Set(const char* key, const std::string& value);

    void Set(const char* key, const double& value);

    void Set(const char* key, float value);

    bool SuccessLoadConfig() const
    {
        return successLoadConfig_;
    }

private:
    std::map<std::string, std::string> stringElem_;
    std::map<std::string, double> doubleElem_;
    std::map<std::string, float> floatElem_;
    std::map<std::string, bool> boolElem_;

    const std::string nullString_;
    const double zeroDouble_;

    bool successLoadConfig_;

    const std::string& DoGetString(const char* key) const;

    const double& DoGetDouble(const char* key) const;

    float DoGetFloat(const char* key) const;

    bool DoGetBool(const char* key) const;

    void DoSet(const char* key, bool value);

    bool Load();

    template <typename T>
    void SaveElem(T& orig, std::ostream& os);

    void Insert(std::map<std::string, std::string>& orig, const char* key, std::map<std::string, std::string>& dest, const char* initialValue);

    void Insert(std::map<std::string, std::string>& orig, const char* key, std::map<std::string, double>& dest, double initialValue);

    void Insert(std::map<std::string, std::string>& orig, const char* key, std::map<std::string, float>& dest, float initialValue);

    void Insert(std::map<std::string, std::string>& orig, const char* key, std::map<std::string, bool>& dest, bool initialValue);
};

inline const double& Preference::GetDouble(const char* key) const
{
    return DoGetDouble(key);
}

inline float Preference::GetFloat(const char* key) const
{
    return DoGetFloat(key);
}

inline bool Preference::GetBool(const char* key) const
{
    return DoGetBool(key);
}

inline const std::string& EffectorPreference::GetString(const char* key) const
{
    return DoGetString(key);
}

inline void EffectorPreference::Set(const char* key, bool value)
{
    DoSet(key, value);
}

#endif
