/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file audiobuffer.cc
/// @brief Class for audio buffering source file.

#include "../include/audiobuffer.h"
#include "../include/preference.h"
#include "../include/audio.h"
#include "../include/misc.h"

template <class AudioPlayer>
Audio::AudioBuffer<AudioPlayer>::AudioBuffer(const EffectorPreference& pref, AudioPlayer& audio)
    : playLoopStart_(false),
      playLoopStop_(false),
      playLoopPause_(false),
      skipBufferFlush_(false),
      singleSampleSize_(0),
      bufferSize_(0),
      bufferCount_(0),
      sampleRate_(0),
      channels_(0),
      previousSeconds_(0),
      currentSeconds_(0.0f),
      secondsPerOutputBuffer_(0.0f),
      playLoopThread_(0),
      pref_(pref),
      audio_(audio)
{}

template <class AudioPlayer>
void Audio::AudioBuffer<AudioPlayer>::Start(unsigned int bits, unsigned int channels, unsigned int sampleRate)
{
    ResetTimer();
    processor_.Reset();

    if(sampleRate_ != sampleRate) {

        sampleRate_ = sampleRate;

        processor_.Calc(pref_, sampleRate_);
    }

    channels_ = channels;

    singleSampleSize_ = bits / 8 * channels;

    secondsPerOutputBuffer_ = outputBufferSize_;
    secondsPerOutputBuffer_ /= singleSampleSize_ * sampleRate_;

    StartPlayThread();
}

/// @brief Process effector and dithering sample buffer, push to play buffer.
///
/// samples[channels][signal]
///
///
/// ex) samples[0][0] -> left channel at 0, samples[0][3] -> left at channel 3.
///
/// samples[1][0] -> right channel at 0, samples[1][5] -> right channel at 5.
/// @param samples Sample buffer.
/// @param bytes Sample buffer size.
template <class AudioPlayer>
void Audio::AudioBuffer<AudioPlayer>::Push(float** const samples, unsigned int bytes)
{
    bufferSize_ = singleSampleSize_ * bytes;

    char tmpBuf[bufferSize_];

    if(channels_ == 2) {

        if(pref_.GetEffectUse()) {

            ResetEffectorIfNeed();

            processor_.EffectorStereo(samples, bytes, tmpBuf);
        } else
            processor_.DitherStereo(samples, bytes, tmpBuf);
	
    } else {

        if(pref_.GetEffectUse()) {

            ResetEffectorIfNeed();

            processor_.EffectorMono(samples, bytes, tmpBuf);
        } else
            processor_.DitherMono(samples, bytes, tmpBuf);
    }

    InternalBufferPush(tmpBuf);

    while(playLoopPause_)
        usecsleep(0, 100);
}

/// @brief Process effector and dithering sample buffer, push to play buffer.
///
/// In sample is stereo, even index is left channel, and odd index is right channel.
///
///
/// ex) samples[0] -> left channel at 0, samples[2] -> left channel at 1
///
/// samples[1] -> right channel at 0, samples[3] -> right channel at 1
/// @param samples Sample buffer.
/// @param bytes Sample buffer size.
template <class AudioPlayer>
void Audio::AudioBuffer<AudioPlayer>::Push(double* const samples, unsigned int bytes)
{
    bufferSize_ = singleSampleSize_ * bytes;

    char tmpBuf[bufferSize_];

    if(channels_ == 2) {

        if(pref_.GetEffectUse()) {

            ResetEffectorIfNeed();

            processor_.EffectorStereo(samples, bytes, tmpBuf);
        } else
            processor_.DitherStereo(samples, bytes, tmpBuf);
	
    } else {

        if(pref_.GetEffectUse()) {

            ResetEffectorIfNeed();

            processor_.EffectorMono(samples, bytes, tmpBuf);
        } else
            processor_.DitherMono(samples, bytes, tmpBuf);

    }

    InternalBufferPush(tmpBuf);

    while(playLoopPause_)
        usecsleep(0, 100);
}

template <class AudioPlayer>
void Audio::AudioBuffer<AudioPlayer>::Push(mad_fixed_t* const left, unsigned int size)
{
    bufferSize_ = singleSampleSize_ * size;

    char tmpBuf[bufferSize_];

    if(pref_.GetEffectUse()) {

        ResetEffectorIfNeed();

        processor_.EffectorMono(left, size, tmpBuf);
    } else
        processor_.DitherMono(left, size, tmpBuf);

    InternalBufferPush(tmpBuf);

    while(playLoopPause_)
        usecsleep(0, 100);
}

template <class AudioPlayer>
void Audio::AudioBuffer<AudioPlayer>::Push(mad_fixed_t* const left, mad_fixed_t* const right, unsigned int size)
{
    bufferSize_ = singleSampleSize_ * size;

    char tmpBuf[bufferSize_];

    if(pref_.GetEffectUse()) {

        ResetEffectorIfNeed();

        processor_.EffectorStereo(left, right, size, tmpBuf);
    } else
        processor_.DitherStereo(left, right, size, tmpBuf);

    InternalBufferPush(tmpBuf);

    while(playLoopPause_)
        usecsleep(0, 100);
}

template <class AudioPlayer>
void Audio::AudioBuffer<AudioPlayer>::Push(short int* const samples, unsigned int size)
{
    bufferSize_ = singleSampleSize_ * size;

    char tmpBuf[bufferSize_];

    if(channels_ == 2) {

        if(pref_.GetEffectUse()) {

            ResetEffectorIfNeed();

            processor_.EffectorStereo(samples, size, tmpBuf);
        } else
            processor_.DitherStereo(samples, size, tmpBuf);
	
    } else {

        if(pref_.GetEffectUse()) {

            ResetEffectorIfNeed();

            processor_.EffectorMono(samples, size, tmpBuf);
        } else
            processor_.DitherMono(samples, size, tmpBuf);
    }

    InternalBufferPush(tmpBuf);

    while(playLoopPause_)
        usecsleep(0, 100);
}

template <class AudioPlayer>
void Audio::AudioBuffer<AudioPlayer>::Close(bool flush)
{
    if(playLoopStart_) {
        playLoopStop_ = true;
        playLoopCondPush_.broadcast();
        playLoopThread_->join();
        playLoopStop_ = false;
		
        playLoopStart_ = false;	
    }
	    
    if(bufferCount_ != 0 && !skipBufferFlush_ && flush)
        audio_.Play(buffer_, bufferCount_);

    audio_.Close();

    bufferCount_ = 0;
    playLoopPause_ = false;
    skipBufferFlush_ = false;
}

template <class AudioPlayer>
void Audio::AudioBuffer<AudioPlayer>::PauseToggle(bool skipPlayBufferedSample)
{
    if(skipPlayBufferedSample)
        skipBufferFlush_ = true;

    if(playLoopPause_)
        playLoopPause_ = false;
    else if(playLoopStart_)
        playLoopPause_ = true;	
}

/// @brief Play buffered samples.
///
/// Pop samples from play buffer and copy to local buffer. After that
/// play local buffer.
template <class AudioPlayer>
void Audio::AudioBuffer<AudioPlayer>::InternalPlayLoop()
{
    for(;;) {

        {
            Glib::Mutex::Lock lock(playLoopMutex_);

            if(bufferCount_ < outputBufferSize_)
                while(bufferCount_ < 131072 && !playLoopStop_) 
                    playLoopCondPush_.wait(playLoopMutex_);

            if(playLoopStop_)
                break;

            memcpy(outputBuffer_, buffer_, outputBufferSize_);

            bufferCount_ -= outputBufferSize_;

            memcpy(buffer_, buffer_ + outputBufferSize_, bufferCount_);	

            currentSeconds_ += secondsPerOutputBuffer_;
            if(previousSeconds_ != static_cast<long int>(currentSeconds_)) {

                previousSeconds_ = static_cast<long int>(currentSeconds_);

                signalTimerUpdate_();
            }
            
            playLoopCondPop_.signal();
        }

        if(!skipBufferFlush_)
            audio_.Play(outputBuffer_, outputBufferSize_);

        while(playLoopPause_)
            usecsleep(0, 100);
    }
}

/// @brief Push sample buffer to play buffer
///
/// Push sample buffer to play buffer. Sample buffer size is
/// member variable bufferSize_. 
/// @param buf Sample buffer to push.
template <class AudioPlayer>
void Audio::AudioBuffer<AudioPlayer>::InternalBufferPush(const char* buf)
{
    Glib::Mutex::Lock lock(playLoopMutex_);

    memcpy(buffer_ + bufferCount_, buf, bufferSize_);

    bufferCount_ += bufferSize_;

    while(bufferCount_ > limitBufferSize_ && !playLoopStop_) 
        playLoopCondPop_.wait(playLoopMutex_);

    playLoopCondPush_.signal();

}

template <class AudioPlayer>
void Audio::AudioBuffer<AudioPlayer>::ResetEffectorIfNeed()
{
    if(pref_.GetEfxChanged()) {

        pref_.EfxChangedUnset();

        processor_.Update(pref_, sampleRate_);
    }
}
