/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file connector.cc
/// @brief Class for network connector source file.

#include <ace/Codecs.h>

#include "../include/connector.h"

#if defined(ACE_HAS_EXPLICIT_TEMPLATE_INSTANTIATION)

template class ACE_Connector<Client, ACE_SOCK_CONNECTOR>;
template class ACE_Svc_Handler<ACE_SOCK_STREAM, ACE_MT_SYNCH>;

#elif defined(ACE_HAS_TEMPLATE_INSTANTIATION_PRAGMA)

#pragma instantiate ACE_Connector<Client, ACE_SOCK_CONNECTOR>
#pragma instantiate ACE_Svc_Handler<ACE_SOCK_STREAM, ACE_MT_SYNCH>

#endif

/// @brief Namespace for some network utility.
namespace NetUtil {

    const size_t packageInfoLength = ACE_OS::strlen(PACKAGE_NAME) + ACE_OS::strlen(VERSION);

    const char* httpProtocolString = "http://";
    const size_t httpProtocolStringLength = ACE_OS::strlen(httpProtocolString);

    /// @brief Urlencode function.
    ///
    /// dest need to have enough space, maybe 3 * orig + 1.
    void UrlEncode(const unsigned char* orig, unsigned char* dest)
    {
        while(*orig) {

            if(NeedUrlEncode(*orig)) {

                *(dest++) = '%';
                *(dest++) = NetUtil::DecToHex( *orig / 16 );
                *(dest++) = NetUtil::DecToHex( *orig % 16 );

            } else
                *(dest++) = *orig;

            ++orig;
        }
        *(dest) = '\0';
    }
};


Address::Address(const ACE_TCHAR* str)
    : portNumber_(80),
      rawHostName_(0),
      request_(0),
      fullAddress_(0),
      authorizationString_(0)
{
    SplitAddressString(str);
}

Address::~Address()
{
    delete [] authorizationString_;

    delete [] fullAddress_;

    delete [] request_;

    delete [] rawHostName_;
}

/// @brief Split original string.
void Address::SplitAddressString(const ACE_TCHAR* str)
{
    // Process address string.
    //
    // Check http protocol string -> Check authorization string ->
    // Check port number string.
    //
    // Possible types. All types are possbile has authorization string.
    //
    // type 1. http://address
    // type 2. http://address:port
    // type 3. http://address/some/where
    // type 4. http://address:port/some/where
    // type 5. address
    // type 6. address:port
    // type 7. address/some/where
    // type 8. address:port/some/where

    if(NetUtil::HasHttpProtocolString(str))
        str += NetUtil::httpProtocolStringLength;

    ParseAuthorizationString(str);
    ParseRawHostname(str);
    ParseRequest(str);
    ParseFullAddress(str);
}

/// @brief Get author string.
void Address::ParseAuthorizationString(const ACE_TCHAR* str)
{
    const char* atCharPosition = ACE_OS::strchr(str, '@');
    if(0 != atCharPosition) {

        const unsigned int offset = atCharPosition - str;

        for(unsigned int i = 0; i < offset; ++i)
            if(str[i] == '/')
                return;
        
        authorizationString_ = new char[ offset + 1];

        ACE_OS::memcpy(authorizationString_,
                       str,
                       offset);

        authorizationString_[offset] = '\0';
    }
}

/// @brief Get host name.
void Address::ParseRawHostname(const ACE_TCHAR* str)
{
    if(0 != authorizationString_)
        str += ACE_OS::strlen(authorizationString_) + 1;

    const char* endPtr = str;
    while(*endPtr && *endPtr != ':' && *endPtr != '/')
        ++endPtr;

    const unsigned int offset = endPtr - str;
    rawHostName_ = new char[offset + 1];

    ACE_OS::memcpy(rawHostName_, str, offset);

    rawHostName_[offset] = '\0';

    if(!*endPtr || *endPtr == '/')
        return;

    // invalid character will be ignored
    portNumber_ = ACE_OS::strtoul(++endPtr, 0, 10);
}

/// @brief Get request.
void Address::ParseRequest(const ACE_TCHAR* str)
{
    const char* atSlashPosition = ACE_OS::strchr(str, '/');
    if(0 != atSlashPosition) {

        request_ = new char[ ACE_OS::strlen(atSlashPosition) * 3 + 1];

        NetUtil::UrlEncode(ACE_reinterpret_cast(const unsigned char*, atSlashPosition), ACE_reinterpret_cast(unsigned char*, request_));

    } else {

        request_ = new char[2];
        request_[0] = '/';
        request_[1] = '\0';

    }
}

/// @brief Get full address.
void Address::ParseFullAddress(const ACE_TCHAR* str)
{
    if(0 != authorizationString_)
        str += ACE_OS::strlen(authorizationString_) + 1;

    fullAddress_ = new char[ ACE_OS::strlen(str) * 3 + 
                             NetUtil::httpProtocolStringLength + 
                             1 ];

    ACE_OS::memcpy(fullAddress_,
                   NetUtil::httpProtocolString,
                   NetUtil::httpProtocolStringLength);

    NetUtil::UrlEncode(ACE_reinterpret_cast(const unsigned char*, str), ACE_reinterpret_cast(unsigned char*, fullAddress_ + NetUtil::httpProtocolStringLength));
}

HttpRequestStrGenerator::HttpRequestStrGenerator(const Address& address, const ACE_TCHAR* accept)
    : requestStr_(0),
      length_(0),
      encodedAuthorizationString_(0)
{
    length_ = GetHttpMessageSize(address, accept);

    requestStr_ = new char[length_];

    MakeHttpGetMessage(address, requestStr_, length_, accept);
}

void HttpRequestStrGenerator::EncodeAuthorizationString(char* str)
{
    size_t encodedLength = 0;
    ACE_Byte* encodedString = ACE_Base64::encode(ACE_reinterpret_cast(ACE_Byte*, str), ACE_OS::strlen(str), &encodedLength);

    if(0 != encodedString) {

        encodedAuthorizationString_ = new ACE_Byte[ encodedLength + 1];
        ACE_OS::memcpy(encodedAuthorizationString_,
                       encodedString,
                       encodedLength);
        encodedAuthorizationString_[encodedLength] = '\0';
    }
}

size_t HttpRequestStrGenerator::GetHttpMessageSize(const Address& address, const ACE_TCHAR* accept)
{
    if(!address.HasAuthorStr())
        return HttpMessageSize(address, accept);
    else {

        size_t authorSize(ACE_OS::strlen(address.GetAuthorStr()));
        char* authorStr = new char[authorSize + 1];
        ACE_OS::strncpy(authorStr, address.GetAuthorStr(), authorSize);
        authorStr[authorSize] = '\0';

        EncodeAuthorizationString(authorStr);

        delete [] authorStr;

        return HttpMessageSize(address, accept) + HttpAuthorSize();
    }
}

/// @brief Create address object.
/// @param destUrl URL string.
/// @param convertCurrentEncoding Convert string to current encoding or not.
std::auto_ptr<Address> CreateAddress(const std::string& destUrl,
                                     bool convertCurrentEncoding)
{
    if(!convertCurrentEncoding) {

        return std::auto_ptr<Address>(new Address(destUrl.c_str()));

    } else {

        try {
            std::string encoding(Glib::locale_from_utf8(destUrl));
            
            return std::auto_ptr<Address>(new Address(encoding.c_str()));

        } catch(Glib::ConvertError& err) {

            return std::auto_ptr<Address>(new Address(destUrl.c_str()));
        }
    }
}

int FileGet::open(void* p)
{
    ACE_Time_Value iterDelay(2);
    if(FileGetHandler::open(p) == -1)
        return -1;

    this->notifier_.reactor(this->reactor());
    this->msg_queue()->notification_strategy(&this->notifier_);

    this->peer().send(generator_->GetString(), generator_->GetLength());

    return this->reactor()->schedule_timer
        (this, 0, ACE_Time_Value::zero, iterDelay);
}

int FileGet::handle_input(ACE_HANDLE fd)
{
    char buf[1024];
    ssize_t recvCnt = this->peer().recv(buf, sizeof(buf) - 1);

    // reduce these duplicated codes :(

    if(recvCnt > 0)
        return ProcessRecvBuf(buf, recvCnt);
    else if(recvCnt == 0 || ACE_OS::last_error() != EWOULDBLOCK) {

        this->reactor()->end_reactor_event_loop();

        return -1;
    }

    return 0;
}

int FileGet::handle_output(ACE_HANDLE fd)
{
    ACE_Message_Block* mb = 0;
    ACE_Time_Value nowait(ACE_OS::gettimeofday());
    while(this->getq(mb, &nowait) != -1) {

        ssize_t sendCnt = this->peer().send(mb->rd_ptr(), mb->length());
        // some send error at sendCnt is -1
        if(sendCnt != -1)
            mb->rd_ptr(ACE_static_cast(size_t, sendCnt));

        if(mb->length() > 0) {

            this->ungetq(mb);
            break;
        }
        mb->release();
    }

    if(this->msg_queue()->is_empty())
        this->reactor()->cancel_wakeup
            (this, ACE_Event_Handler::WRITE_MASK);
    else
        this->reactor()->schedule_wakeup
            (this, ACE_Event_Handler::WRITE_MASK);

    return 0;
}

int FileGet::handle_timeout(const ACE_Time_Value& currentTime, const void* act)
{
    if(this->iterations_ >= ITERATIONS)
        this->peer().close_writer();

    return 0;
}

int FileGet::ReadHeader(char* buf, ssize_t recvCnt)
{
    char* separator = ACE_reinterpret_cast(char*, ACE_OS::memchr(buf, ' ', recvCnt));
    if(0 == separator) {

        this->reactor()->end_reactor_event_loop();

        return -1;
    }

    if(!ReadStatusCode(separator[1]))
        return -1;

    if(redirect_)
        return GetRedirectionURL(buf, separator, recvCnt);
    else
        return SkipRemainHeader(buf, separator, recvCnt);
}

int FileGet::GetRedirectionURL(char* orig, char* startPos, ssize_t limit)
{
    while(0 != startPos) {

        if(ACE_OS::memcmp(startPos + 2, "Location: ", 10) == 0) {

            char* endLine = ACE_reinterpret_cast(char*, ACE_OS::memchr(startPos + 12, '\r', limit - (startPos - orig) - 12));
            if(0 != endLine) {

                const size_t len = (endLine - startPos) - 12;
                redirectURL_ = new char[len + 1];
                ACE_OS::memcpy(redirectURL_, startPos + 12, len);
                redirectURL_[len] = '\0';

                this->reactor()->end_reactor_event_loop();

                break;
            }
        }

        startPos = ACE_reinterpret_cast(char*, ACE_OS::memchr(startPos + 1, '\r', limit - (startPos - orig) - 1));
    }

    return -1;
}

int FileGet::SkipRemainHeader(char* orig, char* startPos, ssize_t limit)
{
    while(0 != startPos) {

        if(ACE_OS::memcmp(startPos, "\r\n\r\n", 4) == 0) {

            recvHeader_ = true;

            ACE_OS::fwrite(startPos + 4,
                           1,
                           limit - 4 - (startPos - orig),
                           outfile_);

            break;
        }

        startPos = ACE_reinterpret_cast(char*, ACE_OS::memchr(startPos + 1, '\r', limit - (startPos - orig) - 1));
    }

    return 0;
}
