/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player. 
//
// Orignal widget first appear in 'Air' project(http://air.kldp.net),
// and this is compact type.
//
//
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file dirbrowsewin.cc
/// @brief Class for directory browse window.

#include <vector>
#include <algorithm>

#include <sys/types.h>
#include <unistd.h>
#include <dirent.h>

#include <libintl.h>

#include "../include/dirbrowsewin.h"

std::string DirBrowseWnd::selectedDirectory_;

/// @brief Default constructor of DirBrowseWnd.
///
/// Window border 12 and empty initial directory.
DirBrowseWnd::DirBrowseWnd()
    : treeModel_(Gtk::TreeStore::create(column_))
{
    scrolledWindow_.set_border_width(12);

    Init();
}

/// @brief Constructor of DirBrowseWnd.
/// @param border Window border width.
/// @param initialDirectory Directory to expand instead current directory at initial.
DirBrowseWnd::DirBrowseWnd(int border, const std::string& initialDirectory)
    : treeModel_(Gtk::TreeStore::create(column_))
{
    selectedDirectory_ = initialDirectory;

    scrolledWindow_.set_border_width(border);

    Init();
}

/// @brief Get selected directory full path from current cursor.
const std::string& DirBrowseWnd::SelectedDirectoryFromIter()
{
    if(GetIterFromCursor())
        selectedDirectory_ = (*GetIterFromCursor())[column_.colPath_];

    return selectedDirectory_;
}

void DirBrowseWnd::Init()
{
    scrolledWindow_.add(treeView_);
    scrolledWindow_.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
 
    treeView_.set_model(treeModel_);
    treeView_.append_column(gettext("Path"), column_.colName_);

    Gtk::Menu_Helpers::MenuList& popupMenuList = menuPopup_.items();
    popupMenuList.push_back( Gtk::Menu_Helpers::StockMenuElem(Gtk::Stock::REFRESH, sigc::mem_fun(*this, &DirBrowseWnd::OnMenuRefresh) ));

    ConnectSignals();

    AddRootDirectory();
}

void DirBrowseWnd::ConnectSignals()
{
    scrolledWindow_.signal_show().connect(sigc::mem_fun(*this, &DirBrowseWnd::OnShow));
    treeView_.signal_row_activated().connect(sigc::mem_fun(*this, &DirBrowseWnd::OnRowActivated));
    treeView_.signal_row_expanded().connect(sigc::mem_fun(*this, &DirBrowseWnd::OnRowExpanded));
    treeView_.signal_button_press_event().connect_notify(sigc::mem_fun(*this, &DirBrowseWnd::OnBtnPressEvent));
}

void DirBrowseWnd::OnShow()
{
    ExpandCurrentDir();
}

/// @brief Callback function used when button is pressed.
///
/// When button is pressed, if button is 3 button, pop up menu.
void DirBrowseWnd::OnBtnPressEvent(GdkEventButton* ev)
{
    if(IsThirdButtonPress(ev))
        menuPopup_.popup(ev->button, ev->time);
}

/// @brief Callback function used when row is activated.
///
/// When row is selected(double click or else ...), read a full path at row.
void DirBrowseWnd::OnRowActivated(const Gtk::TreeModel::Path& path, Gtk::TreeViewColumn* column)
{
    if(GetIterFrom(path)) {

        selectedDirectory_ = (*GetIterFrom(path))[column_.colPath_];

        signalSelected_.emit();
    }
}

/// @brief Callback function used when row is expanded
void DirBrowseWnd::OnRowExpanded(const Gtk::TreeModel::iterator& iter, const Gtk::TreeModel::Path& path)
{
    ChangeWatchCursor();

    std::vector<std::string> dirNames;
    bool hasSubDirectory = false;

    GetUnregisteredSubDirectories((*iter)[column_.colPath_], dirNames, hasSubDirectory);

    if(!hasSubDirectory)
        RemoveDummy(iter);

    if(!dirNames.empty())
        AddDirs(dirNames, iter);

    ChangeMainCursor();
}

/// @brief Expand tree to current(initial) directory.
void DirBrowseWnd::ExpandCurrentDir()
{
    if(!ValidSelectedDirectory())
        selectedDirectory_ = Glib::get_current_dir();

    ExpandFirstRow();

    ScrollToPath(ExpandSubRows());
}


/// @brief Callback funtion used when refresh menu is selected.
///
/// Swap external container to empty one and expand tree to current(initial) directory.
void DirBrowseWnd::OnMenuRefresh()
{
    Gtk::TreeModel::iterator iter(treeModel_->children().begin());

    while(iter)
        iter = treeModel_->erase(iter);

    ResetPathSet();

    AddRootDirectory();

    ExpandCurrentDir();
}

/// @brief Get unregistered sub directoy list.
void DirBrowseWnd::GetUnregisteredSubDirectories(const std::string& currentPath, std::vector<std::string>& dest, bool& hasSubDir)
{
    DIR* const currentDir = opendir(currentPath.c_str());
    dirent* dirEntry = 0;
    std::string fullPath;

    while((dirEntry = readdir(currentDir)) != 0) {

        if(dirEntry->d_name[0] != '.') {

            fullPath = currentPath;
            fullPath += '/';
            fullPath += dirEntry->d_name;

            if(IsDirectory(fullPath)) {

                hasSubDir = true;
                if(!pathInStore_.count(fullPath))
                    dest.push_back(dirEntry->d_name);
            }
        }
    }
    closedir(currentDir); 
}

void DirBrowseWnd::InsertName(Gtk::TreeModel::Row child, const std::string& name, const std::string& fullpath)
{
    pathInStore_.insert(fullpath);
    FillRow(child, name, fullpath);
}

Gtk::TreeModel::iterator DirBrowseWnd::GetIterFromCursor()
{
    Gtk::TreeModel::Path currentPath;
    Gtk::TreeView::Column* currentColumn;

    treeView_.get_cursor(currentPath, currentColumn);

    return treeModel_->get_iter(currentPath);
}

void DirBrowseWnd::AddDirs(std::vector<std::string>& dirNames, const Gtk::TreeModel::iterator& iter)
{
    std::sort(dirNames.begin(), dirNames.end());

    std::vector<std::string>::const_iterator each(dirNames.begin());
    const std::vector<std::string>::const_iterator end(dirNames.end());

    std::string parentPath((*iter)[column_.colPath_]);
    if(parentPath != "/")
        parentPath += '/';

    InsertName(GetDummyRow(iter), *each, parentPath + *each);
    ++each;

    while(each != end) {

        InsertName(GetNewRow(iter), *each, parentPath + *each);
        ++each;
    }
}

Gtk::TreePath DirBrowseWnd::ExpandSubRows()
{

    const Glib::ustring convertedSelectedDirectory(StringToUTF8::Instance().Convert(selectedDirectory_));

    Glib::ustring::const_iterator pathBegin(convertedSelectedDirectory.begin());
    Glib::ustring::const_iterator slash;
    const Glib::ustring::const_iterator pathEnd(convertedSelectedDirectory.end());

    Glib::ustring compareName;

    Gtk::TreeNodeChildren child(treeModel_->get_iter("0")->children());

    Gtk::TreeNodeChildren::iterator parent;

    do {

        if(FindSubDirName(pathBegin, slash, pathEnd)) {

            compareName = Glib::ustring(pathBegin, slash);

            for(Gtk::TreeNodeChildren::iterator iter = child.begin();
                iter != child.end();
                ++iter) {

                if(iter->get_value(column_.colName_) == compareName) {

                    parent = iter;
                    treeView_.expand_row(Gtk::TreePath(iter), false);
                }

            }

            child = parent->children();
        }

        pathBegin = slash;

    } while(pathBegin != pathEnd);

    return Gtk::TreePath(parent);
}
