/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file effect.cc
/// @brief Effector functions source file.

#include <vector>

#include "../include/effect.h"
#include "../include/preference.h"

template <>
void Effect::BiquadCoeff<Effect::LowShelf>::Calc(int sampleRate, int freq, float gain, double Q)
{
    const double omega = 2.0 * M_PI * freq / sampleRate;
		
    const double sn = sin(omega);
    const double cs = cos(omega);

    const double a = pow(10.0, gain * 0.025);
		
    const double beta = sqrt((a * a + 1.0) / Q - pow((a - 1.0), 2.0));

    const double norm = 1.0 / (a + 1.0 + (a - 1.0) * cs + beta * sn);

    a0_ = (a + 1.0 - (a - 1.0) * cs + beta * sn) * a * norm;
		
    a1_ = (a - 1.0 - (a + 1.0) * cs) * 2.0 * a * norm;

    a2_ = (a + 1.0 - (a - 1.0) * cs - beta * sn) * a * norm;

    b1_ = (a - 1.0 + (a + 1.0) * cs) * -2.0 * norm;

    b2_ = (a + 1.0 + (a - 1.0) * cs - beta * sn) * norm;
}

template <>
void Effect::BiquadCoeff<Effect::HighShelf>::Calc(int sampleRate, int freq, float gain, double Q)
{
    const double omega = 2.0 * M_PI * freq / sampleRate;	

    const double sn = sin(omega);
    const double cs = cos(omega);

    const double a = pow(10.0, gain * 0.025);

    const double beta = sqrt((a * a + 1.0) / Q - pow((a - 1.0), 2.0));
		
    const double norm = 1.0 / (a + 1.0 - (a - 1.0) * cs + beta * sn);

    b1_ = (a - 1.0 - (a + 1.0) * cs) * 2.0 * norm;

    b2_ = (a + 1.0 - (a - 1.0) * cs - beta * sn) * norm;
		
    a0_ = (a + 1.0 + (a - 1.0) * cs + beta * sn) * a * norm;

    a1_ = (a - 1.0 + (a + 1.0) * cs) * -2.0 * a * norm;

    a2_ = (a + 1.0 + (a - 1.0) * cs - beta * sn) * a * norm;
}

template <>
void Effect::BiquadCoeff<Effect::Peak>::Calc(int sampleRate, int freq, float gain, double Q)
{
    const double omega = 2.0 * M_PI * freq / sampleRate;
		
    const double sn = sin(omega);
    const double cs = cos(omega);

    const double a = pow(10.0, gain * 0.025);

    const double alpha = sn / (2.0 * Q);
		
    const double norm = 1.0 / (alpha / a + 1.0);
		
    b1_ = -2.0 * cs * norm;
		
    b2_ = (1.0 - (alpha / a)) * norm;
		
    a0_ = (1.0 + alpha * a) * norm;
		
    a1_ = -2.0 * cs * norm;
		
    a2_ = (1.0 - alpha * a) * norm;
}

template <>
void Effect::BiquadCoeff<Effect::LowPass>::Calc(int sampleRate, int freq, float gain, double Q)
{
    const double omega =  1.0 / tan(M_PI * freq / sampleRate);

    a0_ = 1.0 / (1.0 + (gain + omega) * omega);
		
    a1_ = 2.0 * a0_;
		
    a2_ = a0_;

    b1_ = 2.0 * (1.0 - omega * omega) * a0_;

    b2_ = (1.0 - (gain - omega) * omega) * a0_;
}

template <>
void Effect::BiquadCoeff<Effect::HighPass>::Calc(int sampleRate, int freq, float gain, double Q)
{
    const double omega = tan(M_PI * freq / sampleRate);

    a0_ = 1.0 / (1.0 + (gain + omega) * omega);

    a1_ = -2.0 * a0_;

    a2_ = a0_;

    b1_ = 2.0 * (omega * omega - 1.0) * a0_;

    b2_ = (1.0 - (gain - omega) * omega) * a0_;
}

void Effect::Dummy::DoApply(double& sampleL, double& sampleR)
{}

void Effect::Dummy::DoCalc(const Preference& pref, unsigned int sampleRate)
{}

void Effect::Dummy::DoReset()
{}

void Effect::BassBoost::DoApply(double& sampleL, double& sampleR)
{
    low100_.Apply(sampleL, low100Left_);
    low100_.Apply(sampleR, low100Right_);

    peak40_.Apply(sampleL, peak40Left_);
    peak40_.Apply(sampleR, peak40Right_);
}

void Effect::BassBoost::DoCalc(const Preference& pref, unsigned int sampleRate)
{
    low100_.Calc(sampleRate, 100, 3.0f, 0.60295);
    peak40_.Calc(sampleRate, 40, 9.0f, 1.4f);
}

void Effect::BassBoost::DoReset()
{
    low100Left_.Reset();
    low100Right_.Reset();

    peak40Left_.Reset();
    peak40Right_.Reset();
}

void Effect::BassBoost2::DoApply(double& sampleL, double& sampleR)
{
    low199_.Apply(sampleL, low199Left_);
    low199_.Apply(sampleR, low199Right_);
    
    high1309_.Apply(sampleL, high1309Left_);
    high1309_.Apply(sampleR, high1309Right_);
    
    peak80_.Apply(sampleL, peak80Left_);
    peak80_.Apply(sampleR, peak80Right_);
}

void Effect::BassBoost2::DoCalc(const Preference& pref, unsigned int sampleRate)
{
    peak80_.Calc(sampleRate, 80, 6.5f, 0.76077);
    low199_.Calc(sampleRate, 199, 5.5f, 0.33374);
    high1309_.Calc(sampleRate, 1309, 9.0f, 0.25842);
}

void Effect::BassBoost2::DoReset()
{
    peak80Left_.Reset();
    peak80Right_.Reset();

    low199Left_.Reset();
    low199Right_.Reset();

    high1309Left_.Reset();
    high1309Right_.Reset();
}

void Effect::TrebleBoost::DoApply(double& sampleL, double& sampleR)
{
    high5k_.Apply(sampleL, high5kLeft_);
    high5k_.Apply(sampleR, high5kRight_);
}

void Effect::TrebleBoost::DoCalc(const Preference& pref, unsigned int sampleRate)
{
    high5k_.Calc(sampleRate, 5000, 6.0f, 0.7);
}

void Effect::TrebleBoost::DoReset()
{
    high5kLeft_.Reset();
    high5kRight_.Reset();
}

void Effect::GraphicEQ::DoApply(double& sampleL, double& sampleR)
{
    peak50_.Apply(sampleL, peak50Left_);
    peak50_.Apply(sampleR, peak50Right_);

    peak100_.Apply(sampleL, peak100Left_);
    peak100_.Apply(sampleR, peak100Right_);

    peak200_.Apply(sampleL, peak200Left_);
    peak200_.Apply(sampleR, peak200Right_);

    peak400_.Apply(sampleL, peak400Left_);
    peak400_.Apply(sampleR, peak400Right_);

    peak800_.Apply(sampleL, peak800Left_);
    peak800_.Apply(sampleR, peak800Right_);

    peak1500_.Apply(sampleL, peak1500Left_);
    peak1500_.Apply(sampleR, peak1500Right_);

    peak3k_.Apply(sampleL, peak3kLeft_);
    peak3k_.Apply(sampleR, peak3kRight_);

    peak5k_.Apply(sampleL, peak5kLeft_);
    peak5k_.Apply(sampleR, peak5kRight_);

    peak7k_.Apply(sampleL, peak7kLeft_);
    peak7k_.Apply(sampleR, peak7kRight_);

    peak10k_.Apply(sampleL, peak10kLeft_);
    peak10k_.Apply(sampleR, peak10kRight_);

    peak15k_.Apply(sampleL, peak15kLeft_);
    peak15k_.Apply(sampleR, peak15kRight_);
}

void Effect::GraphicEQ::DoCalc(const Preference& pref, unsigned int sampleRate)
{
    peak50_.Calc(sampleRate, 50, pref.GetFloat("EfxGraphicEQGain50"), 1.0);
    peak100_.Calc(sampleRate, 100, pref.GetFloat("EfxGraphicEQGain100"), 1.0);
    peak200_.Calc(sampleRate, 200, pref.GetFloat("EfxGraphicEQGain200"), 1.0);
    peak400_.Calc(sampleRate, 400, pref.GetFloat("EfxGraphicEQGain400"), 1.0);
    peak800_.Calc(sampleRate, 800, pref.GetFloat("EfxGraphicEQGain800"), 1.0);
    peak1500_.Calc(sampleRate, 1500, pref.GetFloat("EfxGraphicEQGain1.5k"), 1.0);
    peak3k_.Calc(sampleRate, 3000, pref.GetFloat("EfxGraphicEQGain3k"), 1.0);
    peak5k_.Calc(sampleRate, 5000, pref.GetFloat("EfxGraphicEQGain5k"), 1.0);
    peak7k_.Calc(sampleRate, 7000, pref.GetFloat("EfxGraphicEQGain7k"), 1.0);
    peak10k_.Calc(sampleRate, 10000, pref.GetFloat("EfxGraphicEQGain10k"), 1.0);
    peak15k_.Calc(sampleRate, 15000, pref.GetFloat("EfxGraphicEQGain15k"), 1.0);
}

void Effect::GraphicEQ::DoReset()
{
    peak50Left_.Reset();
    peak50Right_.Reset();

    peak100Left_.Reset();
    peak100Right_.Reset();

    peak200Left_.Reset();
    peak200Right_.Reset();

    peak400Left_.Reset();
    peak400Right_.Reset();

    peak800Left_.Reset();
    peak800Right_.Reset();

    peak1500Left_.Reset();
    peak1500Right_.Reset();

    peak3kLeft_.Reset();
    peak3kRight_.Reset();

    peak5kLeft_.Reset();
    peak5kRight_.Reset();

    peak7kLeft_.Reset();
    peak7kRight_.Reset();

    peak10kLeft_.Reset();
    peak10kRight_.Reset();

    peak15kLeft_.Reset();
    peak15kRight_.Reset();
}

void Effect::StereoEnhance::DoApply(double& sampleL, double& sampleR)
{
    const double stereoEnhanceDiffLeft = (sampleL - sampleR) * stereoEnhanceAmount_;
    
    sampleL += stereoEnhanceDiffLeft;
    sampleR -= stereoEnhanceDiffLeft;
}

void Effect::StereoEnhance::DoCalc(const Preference& pref, unsigned int sampleRate)
{
    stereoEnhanceAmount_ = pref.GetDouble("EfxStereoEnhanceAmount") * 0.5;
}

void Effect::StereoEnhance::DoReset()
{}

void Effect::Rhythme::DoApply(double& sampleL, double& sampleR)
{
    if(prevCount_ != 18) {

        prevCount_++;

        return;

    } else { 

        const double avg = sqrt(sampleL * sampleR);

        if(feedRhythme_ < rhythmeAmount_ && feedRhythme_ > -rhythmeAmount_ + 0.2 && prevAvg_[0] < avg) { 

            if(prevAvg_[0] * avg < 0.0)
                feedRhythme_ = 0.0;

            if(prevAvg_[1] < prevAvg_[0])
                feedRhythme_ += 0.023293;
            else 
                feedRhythme_ += 0.0115795;

        } else if(feedRhythme_ < rhythmeAmount_ - 0.2 && feedRhythme_ > -rhythmeAmount_ && prevAvg_[0] > avg) {

            if(prevAvg_[0] * avg < 0.0)
                feedRhythme_ = 0.0;

            if(prevAvg_[1] > prevAvg_[0])
                feedRhythme_ -= 0.0227628;
            else
                feedRhythme_ -= 0.0114469;
        }

        sampleL *= 1.0 + feedRhythme_;
        sampleR *= 1.0 + feedRhythme_;
	
        prevCount_ = 0;
        prevAvg_[1] = prevAvg_[0];
        prevAvg_[0] = avg;

    }
}

void Effect::Rhythme::DoCalc(const Preference& pref, unsigned int sampleRate)
{
    rhythmeAmount_ = pref.GetDouble("EfxRhythmeAmount");
}

void Effect::Rhythme::DoReset()
{
    prevCount_ = 0;
    prevAvg_[0] = 0.0;
    prevAvg_[1] = 0.0;
    feedRhythme_ = 0.0;
}

void Effect::Room::DoApply(double& sampleL, double& sampleR)
{
    Stereo filtered(lowPass_.Return(sampleL, lowPassLeft_), lowPass_.Return(sampleR, lowPassRight_));

    if(prevRoomCount_ > 4)
        prevRoomCount_ = 0;

    sampleL += prevLeftSample_[0] * roomAmount_ * 0.75 + 
        prevLeftSample_[1] * roomAmount_ * 0.45 +
        prevLeftSample_[2] * roomAmountSquare_ * 0.60 + 
        prevLeftSample_[3] * roomAmount_ * 0.15 +
        prevLeftSample_[4] * roomAmountSquare_ * 0.30;
	    
    sampleR += prevRightSample_[0] * roomAmount_ * 0.75 +
        prevRightSample_[1] * roomAmount_ * 0.45 +
        prevRightSample_[2] * roomAmountSquare_ * 0.60 +
        prevRightSample_[3] * roomAmount_ * 0.15 +
        prevRightSample_[4] * roomAmountSquare_ * 0.30;

    prevLeftSample_[prevRoomCount_] = filtered.left;
    prevRightSample_[prevRoomCount_] = filtered.right;

    prevRoomCount_++;
}

void Effect::Room::DoCalc(const Preference& pref, unsigned int sampleRate)
{
    lowPass_.Calc(sampleRate, 5300, 0.75, 0.0);
    roomAmount_ = pref.GetDouble("EfxRoomAmount");
    roomAmountSquare_ = roomAmount_ * roomAmount_;
}

void Effect::Room::DoReset()
{
    prevRoomCount_ = 0;

    lowPassLeft_.Reset();
    lowPassRight_.Reset();

    prevLeftSample_[0] = 0.0;
    prevLeftSample_[1] = 0.0;
    prevLeftSample_[2] = 0.0;
    prevLeftSample_[3] = 0.0;
    prevLeftSample_[4] = 0.0;
    
    prevRightSample_[0] = 0.0;
    prevRightSample_[1] = 0.0;
    prevRightSample_[2] = 0.0;
    prevRightSample_[3] = 0.0;
    prevRightSample_[4] = 0.0;
}

void Effect::Room2::DoApply(double& sampleL, double& sampleR)
{
    if(room2Count_[0] >= 882)
        room2Count_[0] = 0;

    room2Count_[1] = room2Count_[0];
    room2Count_[1] -= 441;
    
    if(room2Count_[1] < 0)
        room2Count_[1] += 882;
	
    sampleL += room2PrevSample_[ room2Count_[1] ] * 0.146997;
    room2PrevSample_[ room2Count_[0] ] = sampleL;		
    
    sampleR += room2PrevSample_[ room2Count_[1] + 882 ] * 0.146997;
    room2PrevSample_[ room2Count_[0] + 882 ] = sampleR;		

    ++room2Count_[0];

    low279_.Apply(sampleL, low279Left_);
    low279_.Apply(sampleL, low279Right_);

    low62_.Apply(sampleL, low62Left_);
    low62_.Apply(sampleR, low62Right_);

    high13545_.Apply(sampleL, high13545Left_);
    high13545_.Apply(sampleR, high13545Right_);

    sampleL *= 0.70795;
    sampleR *= 0.70795;

    Stereo cross(sampleR, sampleL);
    
    low852_.Apply(cross.left, low852Left_);
    low852_.Apply(cross.right, low852Right_);

    low35_.Apply(cross.left, low35Left_);
    low35_.Apply(cross.right, low35Right_);

    high3713_.Apply(cross.left, high3713Left_);
    high3713_.Apply(cross.right, high3713Right_);

    cross.left *= room2CrossAmplifyMount_;
    cross.right *= room2CrossAmplifyMount_;

    sampleL += cross.left;
    sampleR += cross.right;
}

void Effect::Room2::DoCalc(const Preference& pref, unsigned int sampleRate)
{
    low62_.Calc(sampleRate, 62, -8.0f, 0.51208);
    low279_.Calc(sampleRate, 279, 4.0f, 0.39539);
    high13545_.Calc(sampleRate, 13545, 4.0f, 0.37929);

    low35_.Calc(sampleRate, 35, -3.0f, 0.52443);
    low852_.Calc(sampleRate, 852, 8.0f + (10.0 * pref.GetDouble("EfxRoom2Amount")), 0.37624);
    high3713_.Calc(sampleRate, 3713, 12.0f, 0.39159);

    room2CrossAmplifyMount_ = pow(10.0, (pref.GetDouble("EfxRoom2Amount") * -10.0 - 15.0) * 0.05);
}

void Effect::Room2::DoReset()
{
    room2Count_[0] = 0;
    room2Count_[1] = 0;

    for(unsigned int i = 0; i < 1764; ++i) {

        room2PrevSample_[i] = 0.0;
    }

    low35Left_.Reset();
    low35Right_.Reset();

    low62Left_.Reset();
    low62Right_.Reset();

    low279Left_.Reset();
    low279Right_.Reset();

    low852Left_.Reset();
    low852Right_.Reset();

    high3713Left_.Reset();
    high3713Right_.Reset();

    high13545Left_.Reset();
    high13545Right_.Reset();
}

void Effect::Harmony::DoApply(double& sampleL, double& sampleR)
{
    Stereo filtered(highPass_.Return(sampleL, highPassLeft_), highPass_.Return(sampleR, highPassRight_));
    
    if(prevHarmonyCount_ > 2)
        prevHarmonyCount_ = 0;

    prevLeftSample_[prevHarmonyCount_] = filtered.left;
    prevRightSample_[prevHarmonyCount_] = filtered.right;
    
    prevHarmonyCount_++;

    const double sampleMultiple = sampleL * sampleR;

    if(sampleMultiple > 0.5625 || sampleMultiple < -0.5625) {

        sampleL += prevLeftSample_[2] * harmonyAmountSquare_ * 0.1 +
            prevLeftSample_[1] * harmonyAmount_ * 0.3 +
            prevLeftSample_[0] * harmonyAmount_ * 0.5;

        sampleR += prevRightSample_[2] * harmonyAmountSquare_ * 0.1 +
            prevRightSample_[1] * harmonyAmount_ * 0.3 +
            prevRightSample_[0] * harmonyAmount_ * 0.5;
    }
}

void Effect::Harmony::DoCalc(const Preference& pref, unsigned int sampleRate)
{
    highPass_.Calc(sampleRate, 10000, 0.5, 0.0);
    harmonyAmount_ = pref.GetDouble("EfxHarmonyAmount");
    harmonyAmountSquare_ = harmonyAmount_ * harmonyAmount_;
}

void Effect::Harmony::DoReset()
{
    prevHarmonyCount_ = 0;

    highPassLeft_.Reset();
    highPassRight_.Reset();

    prevLeftSample_[0] = 0.0;
    prevLeftSample_[1] = 0.0;
    prevLeftSample_[2] = 0.0;

    prevRightSample_[0] = 0.0;
    prevRightSample_[1] = 0.0;
    prevRightSample_[2] = 0.0;
}

void Effect::Distortion::DoApply(double& sampleL, double& sampleR)
{
    sampleL = distortionMutiple_ * atan(sampleL * distortionAmount_);
    sampleR = distortionMutiple_ * atan(sampleR * distortionAmount_);
}

void Effect::Distortion::DoCalc(const Preference& pref, unsigned int sampleRate)
{
    distortionAmount_ = pref.GetDouble("EfxDistortionAmount");

    if(distortionAmount_ > 5.0)
        distortionMutiple_ = 0.2;
    else
        distortionMutiple_ = 1.0 / distortionAmount_;
}

void Effect::Distortion::DoReset()
{}

void Effect::Dither::Reset()
{
    dBuffer_[0] = 0.0;
    dBuffer_[1] = 0.0;
    dBuffer_[2] = 0.0;
    dBuffer_[3] = 0.0;
    dBuffer_[4] = 0.0;
    dBuffer_[5] = 0.0;
    dBuffer_[6] = 0.0;
    dBuffer_[7] = 0.0;
    dBuffer_[8] = 0.0;

    dTState_ = 0.0;

    dPos_ = 8;
}

short int Effect::Dither16Bit::operator()(double& sample, Dither& dither)
{
    Dithering(sample, dither);

    if(ditherResult_ > 32767.0)
        return 32767;
    else if(ditherResult_ < -32768.0)
        return -32768;
    else
        return static_cast<short int>(ditherResult_);
}

void Effect::Dither16Bit::Convert(double& sample, Dither& dither, short int* out)
{
    Dithering(sample, dither);

    if(ditherResult_ > 32767.0)
        *out = 32767;
    else if(ditherResult_ < -32768.0)
        *out = -32768;
    else
        *out = static_cast<short int>(ditherResult_);
}

void Effect::Clip::UpLimit(double& sample)
{
    if(sample > peakSample_) {

        peakSample_ = sample;

        feedGain_ = pow(10.0, -log((peakSample_ - 1.0) * 0.5 + 1.0));
    }

    sample = 1.0;
}

void Effect::Clip::DownLimit(double& sample)
{
    if(sample < -peakSample_) {

        peakSample_ = -sample;
	    
        feedGain_ = pow(10.0, -log((peakSample_ - 1.0) * 0.5 + 1.0));

    }

    sample = -1.0;
}

void Effect::Processor::EffectorStereo(float** const samples, unsigned int size, char* output)
{
    for(unsigned int i = 0; i < size; ++i) {

        double left = samples[0][i];
        double right = samples[1][i];

        amplify_.ApplyFeedGain(left, right);

        ApplyEffector(left, right);

        ApplyDither(left, right, output);
    }
}

void Effect::Processor::EffectorStereo(double* const samples, unsigned int size, char* output)
{
    unsigned int left = 0;
    unsigned int right = 0;

    for(unsigned int i = 0; i < size; ++i) {

        left = i * 2;
        right = left + 1;

        amplify_.ApplyFeedGain(samples[left], samples[right]);

        ApplyEffector(samples[left], samples[right]);

        ApplyDither(samples[left], samples[right], output);
    }
}

void Effect::Processor::EffectorStereo(mad_fixed_t* const left, mad_fixed_t* const right, unsigned int size, char* output)
{
    for(unsigned int i = 0; i < size; ++i) {

        double tempLeft = mad_f_todouble(left[i]);
        double tempRight = mad_f_todouble(right[i]);

        amplify_.ApplyFeedGain(tempLeft, tempRight);

        ApplyEffector(tempLeft, tempRight);

        ApplyDither(tempLeft, tempRight, output);
    }
}

void Effect::Processor::EffectorStereo(short int* const samples, unsigned int size, char* output)
{
    for(unsigned int i = 0; i < size; ++i) {

        double tempLeft = (samples[i * 2] + 0.5) / 32767.5;

        double tempRight = (samples[i * 2 + 1] + 0.5) / 32767.5;

        amplify_.ApplyFeedGain(tempLeft, tempRight);

        ApplyEffector(tempLeft, tempRight);

        ApplyDither(tempLeft, tempRight, output);
    }
}

void Effect::Processor::DitherStereo(float** const samples, unsigned int size, char* output)
{
    for(unsigned int i = 0; i < size; ++i) {

        double left = samples[0][i];
        ApplyDither(left, ditherLeft_, output);

        double right = samples[1][i];
        ApplyDither(right, ditherRight_, output);
    }
}

void Effect::Processor::DitherStereo(double* const samples, unsigned int size, char* output)
{
    unsigned int index = 0;
    for(unsigned int i = 0; i < size; ++i) {

        ApplyDither(samples[index++], ditherLeft_, output);
        ApplyDither(samples[index++], ditherRight_, output);
    }
}

void Effect::Processor::DitherStereo(mad_fixed_t* const left, mad_fixed_t* const right, unsigned int size, char* output)
{
    for(unsigned int i = 0; i < size; ++i) {

        double tempLeft = mad_f_todouble(left[i]);
        double tempRight = mad_f_todouble(right[i]);

        ApplyDither(tempLeft, tempRight, output);
    }
}

void Effect::Processor::DitherStereo(short int* const samples, unsigned int size, char* output)
{
    for(unsigned int i = 0; i < size; ++i) {

        shortIntToChar_(samples[i * 2], output);
        output += 2;
        shortIntToChar_(samples[i * 2 + 1], output);
        output += 2;
    }
}

void Effect::Processor::EffectorMono(double* const samples, unsigned int size, char* output)
{
    unsigned int left = 0;

    for(unsigned int i = 0; i < size; ++i) {

        left = i + i;
        double temp = 0.0;

        amplify_.ApplyFeedGain(samples[left], temp);

        ApplyEffector(samples[left], temp);

        ApplyDither(samples[left], ditherLeft_, output);
    }
}

void Effect::Processor::EffectorMono(float** const samples, unsigned int size, char* output)
{
    for(unsigned int i = 0; i < size; ++i) {

        double left = samples[0][i];
        double right = 0.0;

        amplify_.ApplyFeedGain(left, right);

        ApplyEffector(left, right);

        ApplyDither(left, ditherLeft_, output);
    }
}

void Effect::Processor::EffectorMono(mad_fixed_t* const left, unsigned int size, char* output)
{
    for(unsigned int i = 0; i < size; ++i) {

        double tempLeft = mad_f_todouble(left[i]);
        double tempRight = 0.0;

        amplify_.ApplyFeedGain(tempLeft, tempRight);

        ApplyEffector(tempLeft, tempRight);

        ApplyDither(tempLeft, ditherLeft_, output);
    }
}

void Effect::Processor::EffectorMono(short int* const samples, unsigned int size, char* output)
{
    for(unsigned int i = 0; i < size; ++i) {

        double tempLeft = (samples[i * 2] + 0.5) / 32767.5;
        
        double tempRight = 0.0;

        amplify_.ApplyFeedGain(tempLeft, tempRight);

        ApplyEffector(tempLeft, tempRight);

        ApplyDither(tempLeft, ditherLeft_, output);
    }
}

void Effect::Processor::DitherMono(float** const samples, unsigned int size, char* output)
{
    for(unsigned int i = 0; i < size; ++i) {

        double temp = samples[0][i];
        ApplyDither(temp, ditherLeft_, output);
    }
}

void Effect::Processor::DitherMono(double* const samples, unsigned int size, char* output)
{
    for(unsigned int i = 0; i < size; ++i)
        ApplyDither(samples[i * 2], ditherLeft_, output);
}

void Effect::Processor::DitherMono(mad_fixed_t* const left, unsigned int size, char* output)
{
    for(unsigned int i = 0; i < size; ++i) {

        double tempLeft = mad_f_todouble(left[i]);

        ApplyDither(tempLeft, ditherLeft_, output);
    }
}

void Effect::Processor::DitherMono(short int* const samples, unsigned int size, char* output)
{
    for(unsigned int i = 0; i < size; ++i) {

        shortIntToChar_(samples[i * 2], output);
        output += 2;
    }
}

void Effect::Processor::Update(const Preference& pref, unsigned int sampleRate)
{
    RemoveAll();

    if(pref.GetBool("EfxBassBoost"))
        PutEffector(pref, sampleRate, EffectorCreator::BASSBOOST);

    if(pref.GetBool("EfxBassBoost2"))
        PutEffector(pref, sampleRate, EffectorCreator::BASSBOOST2);

    if(pref.GetBool("EfxTrebleBoost"))
        PutEffector(pref, sampleRate, EffectorCreator::TREBLEBOOST);
 
    if(pref.GetBool("EfxGraphiceq"))
        PutEffector(pref, sampleRate, EffectorCreator::GRAPHICEQ);
 
    if(pref.GetBool("EfxStereoEnhance"))
        PutEffector(pref, sampleRate, EffectorCreator::STEREOENHANCE);
 
    if(pref.GetBool("EfxRhythme"))
        PutEffector(pref, sampleRate, EffectorCreator::RHYTHME);

    if(pref.GetBool("EfxRoom"))
        PutEffector(pref, sampleRate, EffectorCreator::ROOM);

    if(pref.GetBool("EfxRoom2"))
        PutEffector(pref, sampleRate, EffectorCreator::ROOM2);

    if(pref.GetBool("EfxHarmony"))
        PutEffector(pref, sampleRate, EffectorCreator::HARMONY);

    if(pref.GetBool("EfxDistortion"))
        PutEffector(pref, sampleRate, EffectorCreator::DISTORTION);

    iterBegin_ = effectors_.begin();
    iterEnd_ = effectors_.end();
}

void Effect::Processor::Reset()
{
    for(std::vector<Effector*>::iterator iter = iterBegin_;
        iter != iterEnd_;
        ++iter) {

        (*iter)->Reset();

    }

    amplify_.Reset();

    ditherLeft_.Reset();
    ditherRight_.Reset();
}

void Effect::Processor::Calc(const Preference& pref, unsigned int sampleRate)
{
    for(std::vector<Effector*>::iterator iter = iterBegin_;
        iter != iterEnd_;
        ++iter) {

        (*iter)->Calc(pref, sampleRate);

    }
}

void Effect::Processor::DeleteEffectors()
{
    for(std::vector<Effector*>::iterator iter = iterBegin_;
        iter != iterEnd_;
        ++iter) {

        delete *iter;

    }
}

void Effect::Processor::RemoveAll()
{
    DeleteEffectors();

    std::vector<Effector*>().swap(effectors_);
}

void Effect::Processor::PutEffector(const Preference& pref, unsigned int sampleRate, unsigned int type)
{
    Effector* effector = effectorCreator_.Create(type);

    effector->Reset();
    effector->Calc(pref, sampleRate);
    effectors_.push_back(effector);
}

void Effect::Processor::ApplyEffector(double& left, double& right)
{
    std::vector<Effector*>::iterator iter = iterBegin_;
    while(iter != iterEnd_) {

        (*iter++)->Apply(left, right);

        amplify_.ApplyClip(left, right);
    }
}
