/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file mainmanager.cc
/// @brief Class for main program control mediator source file.

#include <map>

#include <libintl.h>

#include "../include/mainmanager.h"
#include "../include/audiobuffer.h"
#include "../include/playcontroller.h"
#include "../include/playlist.h"
#include "../include/dirplaywin.h"
#include "../include/mainwin.h"
#include "../include/prefwin.h"
#include "../include/aboutdlg.h"

#include "../include/dirbrowsedlg.h"

#ifdef ENABLE_HTTP
#include "../include/urldlg.h"
#endif

#include "../include/audio.h"
#include "../include/mixer.h"

#include "../include/gtkmmutil.h"

#include "../include/misc.h"

#include "audiobuffer.cc"
template class Audio::AudioBuffer<Libao::AudioDriver>;

#include "playcontroller.cc"
template class PlayControllerImpl<SaguaAudioBuffer>;

/// @brief Constructor of MainManager.
/// @param arguments Command line arguments.
MainManagerImpl::MainManagerImpl()
    : isCharSetUTF8_(Glib::get_charset()),
      command_(sigc::mem_fun(*this, &MainManager::MainWndOpen),
               sigc::mem_fun(*this, &MainManager::Prev),
               sigc::mem_fun(*this, &MainManager::PlayStart),
               sigc::mem_fun(*this, &MainManager::PauseToggle),
               sigc::mem_fun(*this, &MainManager::Stop),
               sigc::mem_fun(*this, &MainManager::Next),
               sigc::mem_fun(*this, &MainManager::VolumnIncrease),
               sigc::mem_fun(*this, &MainManager::VolumnDecrease),
               sigc::mem_fun(*this, &MainManager::TrayIconFirstBtnClickAction),
               sigc::mem_fun(*this, &MainManager::FileInfo),
               sigc::mem_fun(*this, &MainManager::About),
               sigc::mem_fun(*this, &MainManager::Quit)),
      trayIcon_(command_),
      audio_(Libao::AudioDriver::Instance()),
      audioBuffer_(new SaguaAudioBuffer(preference_, audio_)),
      mixer_(new Audio::OSSMixer(preference_.GetString("MixerDev"))),
      playController_(new PlayControllerImpl<SaguaAudioBuffer>(*audioBuffer_, *this)),
      playlist_(new PlaylistImpl(*this, *playController_, lastState_.GetWindowHeight(), 196)),
      dirPlayWnd_(new DirPlayWndImpl(*playlist_, lastState_.GetLastBrowseDirectory())),
      mainWnd_(new MainWndImpl(*this, *this, *playlist_, *this, *dirPlayWnd_)),
      fileChooserDlg_(mainWnd_->GetWidget(), gettext("Choose files to add"))
{
    SetFileChooserDlgProperty();

    SetTrayIconAction(preference_.GetString("GeneralTrayIconAction"));

    SetDriver();

    signalAudioDeviceOpenFailed_.connect(sigc::mem_fun(*mainWnd_, &MainWnd::ShowAudioOpenFailDlg));

    audioBuffer_->ConnectSignalTimerUpdate(sigc::mem_fun(*this, &MainManager::TimeUpdate));

    playController_->SignalSetDriver().connect(sigc::mem_fun(*this, &MainManager::SetDriver));

#ifdef ENABLE_HTTP
    playlist_->SignalConvertUrlEncoding().connect(sigc::mem_fun(*this, &SettingInterface::GetConvertUrlToCurrentEncoding));
#endif

    mainWnd_->SetPlayMode();
    mainWnd_->SetRepeatCurrent();

    mainWnd_->GetWidget().show();

    if(!preference_.SuccessLoadConfig())
        RunPrefWnd();
}

MainManagerImpl::~MainManagerImpl()
{}

/// @brief Run main loop;
void MainManagerImpl::Run()
{
    Gtk::Main::run();
}

SaguaPreference& MainManagerImpl::GetPreference()
{
    return preference_;
}

#ifdef ENABLE_HTTP
/// @brief Convert url to current encoding or not.
bool MainManagerImpl::GetConvertUrlToCurrentEncoding() const
{
    return !isCharSetUTF8_ && preference_.GetBool("GeneralConvertUrlToCurrentEncoding");
}
#endif

/// @brief Set changed Audio driver.
///
/// If player is playing, just stop play.
/// Otherwise set Audio driver from Preference.
void MainManagerImpl::SetChangedDriver()
{
    if(playController_->InPlay())
        playController_->StopForSetDriver();
    else
        SetDriver();
}

/// @brief Device is opened or not.
/// @sa Libao::Audio::CheckOpenDevice
bool MainManagerImpl::CheckOpenDevice()
{
    return audio_.Check();
}

/// @brief Test audio driver for driver name with option.
/// @param driver Driver.
/// @param options Options.
/// @return Success of test.
/// @sa Libao::Audio::TestDriver(const char* const driverName, const std::vector<std::string>& options)
bool MainManagerImpl::TestDriver(const char* driver, const std::map<std::string, std::string>& options)
{
    return audio_.TestDriver(driver, options);
}

/// @brief Toggle pause state.
/// @sa Audio::PauseToggle
void MainManagerImpl::PauseToggle()
{
    audioBuffer_->PauseToggle(false);
}

/// @brief Unpause audio buffer play.
///
/// Unpause audio buffer play and skip play buffered sample.
/// @sa Audio::PauseToggle
void MainManagerImpl::DoUnpause()
{
    audioBuffer_->Unpause();
}

/// @brief Get current play time from audio output.
long int MainManagerImpl::GetCurrentTime()
{
    return audioBuffer_->GetCurrentTime();
}

/// @brief Refresh Mixer widget.
/// @sa MainWnd::RefreshMixer
void MainManagerImpl::Refresh()
{
    mixer_->SetDevice(preference_.GetString("MixerDev"));

    mainWnd_->SyncMixerWidget();
}

/// @brief Get current PCM volumn.
/// @sa Mixer::GetVolumn
double MainManagerImpl::GetVolumn() const
{
    return mixer_->GetVolumn();
}

/// @brief Set PCM volumn.
/// @param volumn Value to set. 0 ~ 100.
/// @sa Mixer::SetVolumn
void MainManagerImpl::SetVolumn(double volumn)
{
    mixer_->SetVolumn(volumn);
}

/// @brief Mixer Volumn increase.
/// @sa MainWnd::VolumnIncrease 
void MainManagerImpl::VolumnIncrease()
{
    Audio::IncreaseVolumn(*mixer_);

    mainWnd_->SyncMixerWidget();
}

/// @brief Mixer Volumn decrease.
/// @sa MainWnd::VolumnDecrease
void MainManagerImpl::VolumnDecrease()
{
    Audio::DecreaseVolumn(*mixer_);

    mainWnd_->SyncMixerWidget();
}

bool MainManagerImpl::InPlay()
{
    return playController_->InPlay();
}

/// @brief Play previous position file at playlist.
/// @sa MainWnd::OnBtnPrev
void MainManagerImpl::Prev()
{
    playController_->Prev();
}

/// @brief Play file.
/// @sa MainWnd::OnBtnPlay
void MainManagerImpl::PlayStart()
{
    playController_->Play();
}

/// @brief Stop play.
/// @sa MainWnd::OnBtnStop
void MainManagerImpl::Stop()
{
    playController_->Stop();
}

/// @brief Play next position file at playlist.
/// @sa MainWnd::OnBtnNext
void MainManagerImpl::Next()
{
    playController_->Next();
}

void MainManagerImpl::Seek(unsigned long time)
{
    playController_->Seek(time);
}

/// @brief Callback function used when play is finished.
/// @sa MainWnd::FinishPlay
void MainManagerImpl::Finished()
{
    playController_->Finish();
}

/// @brief Callback function used when player need to update timer.
/// @sa MainWnd::TimeUpdate
void MainManagerImpl::TimeUpdate()
{
    mainWnd_->TimeUpdate();
}

void MainManagerImpl::SetNormalPlay()
{
    playController_->SetItemSelector(*playlist_, ItemSelectorCreator::NORMAL);
}

void MainManagerImpl::SetRepeatPlay()
{
    playController_->SetItemSelector(*playlist_, ItemSelectorCreator::REPEAT);
}

void MainManagerImpl::SetRepeatCurrentPlay(bool repeat)
{
    playController_->SetRepeatCurrent(repeat);
}

void MainManagerImpl::SetRandomPlay()
{
    playController_->SetItemSelector(*playlist_, ItemSelectorCreator::RANDOM);
}

void MainManagerImpl::UpdateWndItemRead()
{
    mainWnd_->UpdateItemRead();
}

void MainManagerImpl::UpdateWndPlay()
{
    mainWnd_->UpdatePlay();
}

void MainManagerImpl::UpdateWndStop()
{
    mainWnd_->UpdateStop();
}

void MainManagerImpl::UpdateWndItemRemoved()
{
    mainWnd_->UpdateItemRemoved();
}

/// @brief Show current file information.
/// @sa MainWnd::CurrentFileInfo
void MainManagerImpl::FileInfo()
{
    playlist_->ShowCurrentFileInfoDlg();
}

/// @brief Set tooltips for trayicon.
/// @sa TrayIcon::SetTooltips
void MainManagerImpl::TrayIconSetTooltips(const Glib::ustring& string)
{
    trayIcon_.SetTooltips(string);
}

void MainManagerImpl::SetTrayIconAction()
{
    SetTrayIconAction(preference_.GetString("GeneralTrayIconAction"));
}

/// @brief Callback function when trayicon is pressed by 1 button.
void MainManagerImpl::TrayIconFirstBtnClickAction()
{
    (this->*(currentTrayIconAction_))();
}

/// @brief Notify change of compact mode background change to MainWnd.
void MainManagerImpl::MainWndChangeCompactBgPixmap()
{
    mainWnd_->ChangeCompactBgPixmap();
}

void MainManagerImpl::PushList(const std::vector<std::string>& list)
{
    playlist_->RemoveAll();

    playlist_->AddFiles(list, true);

    playlist_->ForceUpdateToppage();

    PlayStart();
}

sigc::signal1<void, std::vector<std::string>&>& MainManagerImpl::SignalReceived()
{
    return signalReceived_;
}

/// @brief Set MainWnd statusbar text and update it.
/// @sa MainWnd::SetStatusbar
void MainManagerImpl::SetStatusbar(const Glib::ustring& text)
{
    mainWnd_->SetStatusbar(text);
}

/// @brief Set MainWnd statusbar text.
/// @sa MainWnd::SetStatusbarText
void MainManagerImpl::SetStatusbarText(const Glib::ustring& text)
{
    mainWnd_->SetStatusbarText(text);
}

/// @brief Update MainWnd statusbar to stored text.
/// @sa MainWnd::UpdateStatusbar
void MainManagerImpl::UpdateStatusbar()
{
    mainWnd_->UpdateStatusbar();
}

/// @brief Update MainWnd statusbar to stored text and clear at timeout.
/// @sa MainWnd::UpdateStatusbarTimeoutClear
void MainManagerImpl::UpdateStatusbarTimeoutClear()
{
    mainWnd_->UpdateStatusbarTimeoutClear();
}

/// @brief Clear MainWnd statusbar.
/// @sa MainWnd::ClearStatusbar
void MainManagerImpl::ClearStatusbar()
{
    mainWnd_->ClearStatusbar();
}

/// @brief Clear MainWnd statusbar at timeout.
/// @sa MainWnd::TimeoutClearStatusbar
void MainManagerImpl::TimeoutClearStatusbar()
{
    mainWnd_->TimeoutClearStatusbar();
}

LastState& MainManagerImpl::GetLastState()
{
    return lastState_;
}

void MainManagerImpl::FileAdd()
{
    if(Gtk::RESPONSE_OK == fileChooserDlg_.run()) {

        playlist_->AddFiles(fileChooserDlg_.get_filenames(), true);

        playlist_->ForceUpdatePage();
    }

    fileChooserDlg_.hide();
}

void MainManagerImpl::DirAdd()
{
    DirBrowseDlg dirBrowseDlg(mainWnd_->GetWidget());

    if(Gtk::RESPONSE_OK == dirBrowseDlg.Run()) {

        SetMouseCursorWatch(mainWnd_->GetWidget());

        playlist_->AddDirectoryRecursive(dirBrowseDlg.SelectedDirectory());

        RestoreMouseCursor(mainWnd_->GetWidget());

        playlist_->ForceUpdatePage();
    }
}

#ifdef ENABLE_HTTP

void MainManagerImpl::UrlAdd()
{
    UrlDlg urlDlg(mainWnd_->GetWidget());

    if(Gtk::RESPONSE_OK == urlDlg.Run()) {

        const std::string& selectedUrl(urlDlg.SelectedUrl());

        if(selectedUrl.empty())
            return;

        if(HasHttpProtocolString(selectedUrl))
            playlist_->AddUrl(selectedUrl);
        else
            playlist_->AddUrl("http://" + selectedUrl);
    }
}

#endif

void MainManagerImpl::DoSetDriver()
{
    std::map<std::string, std::string> option;
    preference_.GetAudioDriverOption(option);

    audio_.SetDriver(preference_.GetString("AudioDriver").c_str(), option);
}

/// @brief Open Audio device.
/// @param bits Bits per sample.
/// @param channels Number of channels.
/// @param sampleRate Sample rate per sample.
/// @return Success of open.
/// @sa Libao::Audio::OpenDevice
bool MainManagerImpl::OpenDevice(unsigned int bits, unsigned int channels, unsigned int sampleRate)
{
    if(audio_.OpenDevice(bits, channels, sampleRate)) {

        audioBuffer_->Start(bits, channels, sampleRate);

        return true;

    } else {

        Stop();

        signalAudioDeviceOpenFailed_();

        return false;
    }
}

/// @brief Add received file list to playlist.
void MainManagerImpl::SocketReceived()
{
    std::vector<std::string> received;

    signalReceived_(received);

    PushList(received);
}

/// @brief Show MainWnd and raise upon.
void MainManagerImpl::MainWndOpen()
{
    mainWnd_->GetWidget().show();

    mainWnd_->GetWidget().raise();
}

/// @brief Show about dialog.
void MainManagerImpl::About()
{
    AboutDlg aboutDlg(mainWnd_->GetWidget());

    aboutDlg.Run();
}

/// @brief Hide MainWnd and exit loop.
void MainManagerImpl::Quit()
{
    Stop();

    mainWnd_->SaveState();
    mainWnd_->GetWidget().hide();

    Gtk::Main::quit();
}

/// @brief Run PrefWnd.
void MainManagerImpl::RunPrefWnd()
{
    PrefWnd prefWnd(mainWnd_->GetWidget());
    prefWnd.Run();
}

/// @brief Set tray icon action from parameter.
void MainManagerImpl::SetTrayIconAction(const std::string& action)
{
    if(action == "openwin")
        currentTrayIconAction_ = &MainManager::MainWndOpen;

    else if(action == "fileinfo")
        currentTrayIconAction_ = &MainManager::FileInfo;

    else if(action == "stop")
        currentTrayIconAction_ = &MainManager::Stop;

    else if(action == "pause")
        currentTrayIconAction_ = &MainManager::PauseToggle;

    else if(action == "play")
        currentTrayIconAction_ = &MainManager::PlayStart;

    else if(action == "prev")
        currentTrayIconAction_ = &MainManager::Prev;

    else if(action == "next")
        currentTrayIconAction_ = &MainManager::Next;
}

void MainManagerImpl::SetFileChooserDlgProperty()
{
    fileChooserDlg_.set_local_only(true);
    fileChooserDlg_.set_select_multiple(true);

    Gtk::FileFilter filterAll;
    filterAll.set_name(gettext("All files"));
    filterAll.add_pattern("*");
    fileChooserDlg_.add_filter(filterAll);

    Gtk::FileFilter filterPlayable;
    filterPlayable.set_name(gettext("All playable files"));
    filterPlayable.add_pattern("*.[Mm][Pp]1");
    filterPlayable.add_pattern("*.[Mm][Pp]2");
    filterPlayable.add_pattern("*.[Mm][Pp]3");
    filterPlayable.add_pattern("*.[Oo][Gg][Gg]");
    filterPlayable.add_pattern("*.[Ww][Mm][Aa]");
    fileChooserDlg_.add_filter(filterPlayable);

    Gtk::FileFilter filterMpeg;
    filterMpeg.set_name("Mpeg 1 Audio");
    filterMpeg.add_pattern("*.[Mm][Pp]1");
    filterMpeg.add_pattern("*.[Mm][Pp]2");
    filterMpeg.add_pattern("*.[Mm][Pp]3");
    fileChooserDlg_.add_filter(filterMpeg);

    Gtk::FileFilter filterOgg;
    filterOgg.set_name("Ogg vorbis");
    filterOgg.add_pattern("*.[Oo][Gg][Gg]");
    fileChooserDlg_.add_filter(filterOgg);

    Gtk::FileFilter filterWma;
    filterWma.set_name("Window Media Audio");
    filterWma.add_pattern("*.[Ww][Mm][Aa]");
    fileChooserDlg_.add_filter(filterWma);

    Gtk::FileFilter filterList;
    filterList.set_name(gettext("Playlist file"));
    filterList.add_pattern("*.[Mm]3[Uu]");
    filterList.add_pattern("*.[Pp][Ll][Ss]");
    fileChooserDlg_.add_filter(filterList);

    fileChooserDlg_.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
    fileChooserDlg_.add_button(Gtk::Stock::OK, Gtk::RESPONSE_OK);
}
