/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file playcontroller.cc
/// @brief Class for play control source file.

#include "../include/playcontroller.h"

#include "../include/player_mp3.h"
#include "../include/player_ogg.h"
#include "../include/player_wma.h"

#ifdef ENABLE_FAAD2
#include "../include/player_aac.h"
#endif

#ifdef ENABLE_HTTP
#include "../include/player_mp3http.h"
#endif

#include "../include/mainmanager.h"
#include "../include/mainwin.h"
#include "../include/playlist.h"
#include "../include/scopeguard.h"
#include "../include/exception.h"
#include "../include/misc.h"

#include "../include/audiobuffer.h"

#include "player_mp3.cc"
template class MAD::MP3Player<SaguaAudioBuffer>;

#include "player_ogg.cc"
template class VORBIS::OGGPlayer<SaguaAudioBuffer>;

#include "player_wma.cc"
template class FFMPEG::WMAPlayer<SaguaAudioBuffer>;

#ifdef ENABLE_FAAD2
#include "player_aac.cc"
template class FAAD::AACPlayer<SaguaAudioBuffer>;
#endif

#ifdef ENABLE_HTTP
#include "player_mp3http.cc"
template class MAD::MP3HttpPlayer<SaguaAudioBuffer>;
#endif

template <class Buffer>
PlayControllerImpl<Buffer>::PlayControllerImpl(Buffer& buffer, PlayControlDisplay& display)
    : inThread_(false),
      inPlay_(false),
      breakPlay_(false),
      breakPlayForSetDriver_(false),
      repeatCurrent_(false),
      playThread_(0),
      playerFactory_(new PlayerFactory<Buffer>(buffer)),
      buffer_(buffer),
      display_(display)
{
    RegisterPlayerFactory();
}

template <class Buffer>
PlayControllerImpl<Buffer>::~PlayControllerImpl()
{
    if(inThread_) {

        buffer_.Unpause();
        inPlay_ = false;
        player_->Stop();
        playThread_->join();
    }
}

template <class Buffer>
void PlayControllerImpl<Buffer>::Play()
{
    if(currentItem_.empty()) {

        if(!selector_->GetNext(currentItem_))
            return;
    }

    display_.UpdateWndItemRead();

    inPlay_ = true;

    PlayInternal();
}

template <class Buffer>
void PlayControllerImpl<Buffer>::PlayFile(const std::string& name)
{
    currentItem_ = name;

    display_.UpdateWndItemRead();

    Play();
}

template <class Buffer>
void PlayControllerImpl<Buffer>::Next()
{
    if(!selector_->GetNext(currentItem_))
        return;

    display_.UpdateWndItemRead();

    if(inPlay_)
        PlayInternal();
}

template <class Buffer>
void PlayControllerImpl<Buffer>::Prev()
{
    if(!selector_->GetPrev(currentItem_))
        return;

    display_.UpdateWndItemRead();

    if(inPlay_)
        PlayInternal();
}

template <class Buffer>
void PlayControllerImpl<Buffer>::Stop()
{
    StopInternal();
}

template <class Buffer>
void PlayControllerImpl<Buffer>::StopForSetDriver()
{
    breakPlayForSetDriver_ = true;
    StopInternal();
}

template <class Buffer>
void PlayControllerImpl<Buffer>::ItemRemoved()
{
    StopInternal();
    currentItem_ = "";
    display_.UpdateWndItemRemoved();
}

template <class Buffer>
void PlayControllerImpl<Buffer>::Seek(unsigned long time)
{
    player_->Seek(time);
}

template <class Buffer>
void PlayControllerImpl<Buffer>::Finish()
{
    playThread_->join();
    inThread_ = false;

    if(breakPlayForSetDriver_) {

        breakPlayForSetDriver_ = false;
        signalSetDriver_();
    }

    if(inPlay_) {

        if(breakPlay_) {

            breakPlay_ = false;
            PlayInternal();

        } else {

            if(repeatCurrent_)
                PlayInternal();
            else
                NextInternal();
        }

    } else
        display_.UpdateWndStop();
}

template <class Buffer>
bool PlayControllerImpl<Buffer>::IsPlayable(const std::string& filename) const
{
    return playerFactory_->Exist(filename);
}

template <class Buffer>
bool PlayControllerImpl<Buffer>::InPlay() const
{
    return inThread_;
}

template <class Buffer>
void PlayControllerImpl<Buffer>::SetRepeatCurrent(bool repeat)
{
    repeatCurrent_ = repeat;
}

template <class Buffer>
void PlayControllerImpl<Buffer>::PlayInternal()
{
    if(inThread_) {

        buffer_.Unpause();
        breakPlay_ = true;
        player_->Stop();

        return;
    }

    if(IsReadable(currentItem_)) {

        player_ = playerFactory_->Create(currentItem_);

#ifdef ENABLE_HTTP

    } else if(HasHttpProtocolString(currentItem_)) {

        // fix to create at factory!

        try {

            player_ = CreateMP3HttpPlayer(buffer_, currentItem_);

        } catch(const Exception::SaguaException& ex) {

            NextInternal();

            return;
        }

#endif

    } else {

        NextInternal();

        return;
    }

    if(0 == player_.get()) {

        NextInternal();

        return;
    }

    playThread_ = Glib::Thread::create(sigc::mem_fun(*player_, &Player::Start), true);
    inThread_ = true;

    display_.UpdateWndPlay();
}

template <class Buffer>
void PlayControllerImpl<Buffer>::NextInternal()
{
    if(!selector_->GetNext(currentItem_)) {

        GetInternalFailed();

        return;
    }

    display_.UpdateWndItemRead();

    PlayInternal();
}

template <class Buffer>
void PlayControllerImpl<Buffer>::PrevInternal()
{
    if(!selector_->GetPrev(currentItem_)) {

        GetInternalFailed();

        return;
    }

    display_.UpdateWndItemRead();

    PlayInternal();
}

template <class Buffer>
void PlayControllerImpl<Buffer>::StopInternal()
{
    inPlay_ = false;

    if(inThread_) {

        buffer_.Unpause();
        player_->Stop();
    }
}

template <class Buffer>
void PlayControllerImpl<Buffer>::SetItemSelector(GetItem& getItem, unsigned int type)
{
    selector_ = selectorCreator_.Create(type, getItem);
}

template <class Buffer>
void PlayControllerImpl<Buffer>::RegisterPlayerFactory()
{
    // Do you have mpeg 1 layer 1 file? Anyway, i call it mp1.
    playerFactory_->Register("mp1", &CreateMP3Player); 
    playerFactory_->Register("mp2", &CreateMP3Player); 
    playerFactory_->Register("mp3", &CreateMP3Player); 
    playerFactory_->Register("ogg", &CreateOGGPlayer); 
    playerFactory_->Register("wma", &CreateWMAPlayer); 

#ifdef ENABLE_FAAD2
    playerFactory_->Register("mp4", &CreateAACPlayer);
    playerFactory_->Register("m4a", &CreateAACPlayer);
    playerFactory_->Register("aac", &CreateAACPlayer);
#endif
}

template <class Buffer>
void PlayControllerImpl<Buffer>::GetInternalFailed()
{
    inPlay_ = false;
    display_.UpdateWndStop();
}
