/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file player_aac.cc
/// @brief Class for mpeg 4 audio decoding using FAAD2 source file.

#include <cstring>
#include <string>
#include <cstring>

#include "../include/player_aac.h"
#include "../include/aacbuffer.h"
#include "../include/mainmanager.h"
#include "../include/exception.h"
#include "../include/scopeguard.h"

template <class Buffer>
FAAD::AACPlayer<Buffer>::AACPlayer(Buffer& buffer, const std::string& filename)
    : decodeState_(STOP),
      seekSeconds_(0),
      maxChannels_(2),
      frameSize_(1024),
      inFile_(filename.c_str(), "rb"),
      mp4File_(IsMP4File(inFile_.Get())),
      sampleBuffer_(0),
      buffer_(buffer),
      playerInterface_(MainManagerImpl::Instance())
{
    signalFinish_.connect(sigc::mem_fun(playerInterface_, &PlayerInterface::Finished));
}

template <class Buffer>
FAAD::AACPlayer<Buffer>::~AACPlayer()
{}

template <class Buffer>
void FAAD::AACPlayer<Buffer>::DoStart()
{
    decodeState_ = PLAY;

    if(mp4File_) {

        MP4Decode();

    } else {

        AACDecode();

    }

    buffer_.Close(decodeState_ == PLAY);

    signalFinish_();
}

template <class Buffer>
void FAAD::AACPlayer<Buffer>::DoStop()
{
    decodeState_ = STOP;
}

template <class Buffer>
void FAAD::AACPlayer<Buffer>::DoSeek(unsigned long time)
{
    decodeState_ = SEEK;
    seekSeconds_ = time;

    buffer_.SetCurrentTime(time);
}

/// @brief Decode MP4 type file.
/// @todo Need real testing. I have few MP4 encoded file, and works fine with them. But will work with others?
template <class Buffer>
void FAAD::AACPlayer<Buffer>::MP4Decode()
{
    const std::auto_ptr<mp4ff_callback_t> mp4Cb(SetMP4ffCallback(inFile_.Get()));

    faacDecHandle decoder = faacDecOpen();
    ScopeGuard closeDecoder = MakeGuard(faacDecClose, decoder);

    SetDecoderConfig(decoder);

    mp4ff_t* const inFile = mp4ff_open_read(mp4Cb.get());
    if(0 == inFile)
        return;
    ScopeGuard closeMP4ffFile = MakeGuard(mp4ff_close, inFile);

    int track = GetAACTrack(inFile);
    if(track < 0)
        return;

    unsigned char* buffer = 0;
    unsigned int bufferSize = 0;
    mp4ff_get_decoder_config(inFile, track, &buffer, &bufferSize);

    unsigned long sampleRate = 0;
    unsigned char channels = 0;
    if(faacDecInit2(decoder, buffer, bufferSize, &sampleRate, &channels) < 0)
        return;

    // Decoder was set to down matrix, but double check.
    if(static_cast<unsigned int>(channels) > maxChannels_)
        return;

    mp4AudioSpecificConfig mp4ASC;
    if(0 != buffer) {
        if(AudioSpecificConfig(buffer, bufferSize, &mp4ASC) >= 0) {
            if(mp4ASC.frameLengthFlag == 1)
                frameSize_ = 960;
            if(mp4ASC.sbr_present_flag == 1)
                frameSize_ *= 2;	
        }

        free(buffer);
        buffer = 0;
        bufferSize = 0;
    }

    const long numSamples = mp4ff_num_samples(inFile, track);
    int rc = 0;
    faacDecFrameInfo frameInfo;

    if(!playerInterface_.OpenDevice(16, static_cast<unsigned int>(channels), sampleRate))
        return;

    for(long sampleId = 0; sampleId < numSamples; ++sampleId) {
        if(decodeState_ == STOP)
            break;

        if(decodeState_ == SEEK) {

            sampleId = ConvertSecondToCount(seekSeconds_, mp4ASC.samplingFrequency, frameSize_);

            if(sampleId < 0)
                sampleId = 0;

            if(sampleId > numSamples)
                break;

            buffer_.SetCurrentTime(seekSeconds_);

            decodeState_ = PLAY;

        }

        rc = mp4ff_read_sample(inFile, track, sampleId, &buffer, &bufferSize);
        if(rc == 0)
            return;

        sampleBuffer_ = faacDecDecode(decoder, &frameInfo, buffer, bufferSize);

        if(0 != buffer) {

            free(buffer);
            buffer = 0;
        }

        if(frameInfo.error == 0 && frameInfo.samples > 0) 
            buffer_.Push(reinterpret_cast<double*>(sampleBuffer_), frameInfo.samples>>1);

    }
}

/// @brief Decode AAC type file.
/// @todo Need real testing. Implement seek function.
template <class Buffer>
void FAAD::AACPlayer<Buffer>::AACDecode()
{
    AACBuffer aacBuffer(inFile_.Get());

    if(!aacBuffer.InitBuffer(FAAD_MIN_STREAMSIZE * maxChannels_))
        return;

    faacDecHandle decoder = faacDecOpen();
    ScopeGuard closeDecoder = MakeGuard(faacDecClose, decoder);

    SetDecoderConfig(decoder);

    aacBuffer.SkipInfo();

    aacBuffer.FillBuffer();
    unsigned long sampleRate = 0;
    unsigned char channels = 0;
    if(!aacBuffer.InitDecoder(decoder, sampleRate, channels))
        return;

    // Decoder was set to down matrix, but double check.
    if(static_cast<unsigned int>(channels) > maxChannels_)
        return;

    faacDecFrameInfo frameInfo;

    if(!playerInterface_.OpenDevice(16, static_cast<unsigned int>(channels), sampleRate))
        return;

    do {
        if(decodeState_ == STOP)
            break;

        sampleBuffer_ = aacBuffer.Decode(decoder, frameInfo);

        aacBuffer.AdvanceBuffer(frameInfo.bytesconsumed);

        if(frameInfo.error == 0 && frameInfo.samples > 0) 
            buffer_.Push(reinterpret_cast<double*>(sampleBuffer_), frameInfo.samples>>1);

        aacBuffer.FillBuffer();
        if(aacBuffer.IsEmpty())
            sampleBuffer_ = 0;

    } while(sampleBuffer_ != 0);

}

/// @brief Set decoder configuration.
/// @param decoder Decoder.
/// @param defObjectType Set object type.
/// @param outputFormat Library output formats. 
/// @param downMatrix Change multichannels to less one.
/// @param useOldADTSFormat Use old ADTS format.
/// @param dontUpSampleImplicitSBR SBR signalling.
template <class Buffer>
void FAAD::AACPlayer<Buffer>::SetDecoderConfig(faacDecHandle& decoder, const unsigned char& defObjectType, const unsigned char& outputFormat, const unsigned char& downMatrix, const unsigned char& useOldADTSFormat, const unsigned char& dontUpSampleImplicitSBR)
{
    faacDecConfigurationPtr decoderConfig = faacDecGetCurrentConfiguration(decoder);
    decoderConfig->defObjectType = defObjectType;
    decoderConfig->outputFormat = outputFormat;
    decoderConfig->downMatrix = downMatrix;
    decoderConfig->useOldADTSFormat = useOldADTSFormat;
    decoderConfig->dontUpSampleImplicitSBR = dontUpSampleImplicitSBR;
    faacDecSetConfiguration(decoder, decoderConfig);
}

/// @brief Create AACPlayer object.
///
/// Used in PlayerFactory.
template <class Buffer>
std::auto_ptr<Player> CreateAACPlayer(Buffer& buffer, const std::string& filename)
{
    return std::auto_ptr<Player>(new FAAD::AACPlayer<Buffer>(buffer, filename));
}
