/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file player_mp3.cc
/// @brief Class for play mp3 source file.

#include <cmath>

#include "../include/mainmanager.h"
#include "../include/player_mp3.h"
#include "../include/exception.h"
#include "../include/trace.h"

template <class Buffer>
Buffer* MAD::MP3Player<Buffer>::buffer_ = 0;

template <class Buffer>
PlayerInterface* MAD::MP3Player<Buffer>::playerInterface_ = 0;

template <class Buffer>
std::string MAD::MP3Player<Buffer>::filename_;

template <class Buffer>
DecodeState MAD::MP3Player<Buffer>::decodeState_;

template <class Buffer>
mad_timer_t MAD::MP3Player<Buffer>::timer_;

template <class Buffer>
long MAD::MP3Player<Buffer>::seekSeconds_;

/// @brief Constructor of MP3Player.
///
/// Reset static member variables and connect dispatchers to functions.
template <class Buffer>
MAD::MP3Player<Buffer>::MP3Player(Buffer& buffer, const std::string& filename)
{
    filename_ = filename;

    Init();

    if(0 == buffer_)
        buffer_ = &buffer;

    if(0 == playerInterface_)
        playerInterface_ = &MainManagerImpl::Instance();
    
    signalFinish_.connect(sigc::mem_fun(playerInterface_, &PlayerInterface::Finished));
}

/// @brief Initialize MP3Player.
///
/// map file into memory and initialize mad decoder.
/// @throw ErrnoException Exception on init.
template <class Buffer>
void MAD::MP3Player<Buffer>::Init()
{
    DEBUGTRACE("MAD::MP3Player<Buffer>::Init");
    
    if(lstat(filename_.c_str(), &buf_.stat) == -1)
        throw Exception::ErrnoException(errno);
    
    buf_.fd = open(filename_.c_str(), O_RDONLY);
    if(-1 == buf_.fd)
        throw Exception::ErrnoException(errno);
    
    void* tempMap = mmap(0, buf_.stat.st_size, PROT_READ, MAP_SHARED, buf_.fd, 0);	
    if(tempMap == MAP_FAILED) {

        close(buf_.fd);
        buf_.fd = -1;

        throw Exception::ErrnoException(errno);
    }

    buf_.fdm = static_cast<unsigned char*>(tempMap);

    buf_.data = buf_.fdm;

    buf_.length = buf_.stat.st_size;
    buf_.eof = false;

    mad_decoder_init(&decoder_, &buf_, Input, Header, 0, Output, 0, 0); 
}

/// @brief Start player loop
/// 
/// Reset static member variables and start loop. If decodeState is
/// SEEK_REWIND, stop player and seek from begin to wanted position.
///
/// When play is finish, close audio and signal to MainManager.
template <class Buffer>
void MAD::MP3Player<Buffer>::DoStart()
{
    ResetStatic();
    decodeState_ = PLAY;
    mad_decoder_run(&decoder_, MAD_DECODER_MODE_SYNC);
    while(decodeState_ == SEEK_REWIND) {

        if(!Rewind())
            break;
    }

    Clean();

    buffer_->Close(decodeState_ == PLAY);
		
    signalFinish_();
}

/// @brief Set decodeState to STOP
///
/// Set decodeState to STOP. Header function detects STOP and send 
/// MAD_FLOW_STOP to decoder.
template <class Buffer>
void MAD::MP3Player<Buffer>::DoStop()
{
    decodeState_ = STOP;
}

/// @brief Set decodeState to SEEK and set seekSeconds to seconds.
///
/// Set decodeState to SEEK. Header function detects SEEK and
/// judge just SEEK or SEEK_REWIND.
/// @param seconds wanted time to seek in seconds.
template <class Buffer>
void MAD::MP3Player<Buffer>::DoSeek(unsigned long time)
{
    decodeState_ = SEEK;    

    seekSeconds_ = time;

    buffer_->SetCurrentTime(time);
}

/// @brief Clean up decoder.
template <class Buffer>
void MAD::MP3Player<Buffer>::Clean()
{
    mad_decoder_finish(&decoder_);

    if(buf_.fdm)
        munmap(buf_.fdm, buf_.stat.st_size);

    if(!buf_.eof)
        buf_.data = 0;

    if(buf_.data)
        free(buf_.data);

    if(-1 != buf_.fd)
        close(buf_.fd);
}

/// @brief Reset static member variables.
///
/// Almost member variables are static, need to reset them by hands.
template <class Buffer>
void MAD::MP3Player<Buffer>::ResetStatic()
{
    timer_ = mad_timer_zero;
    
    seekSeconds_ = 0;
}

template <class Buffer>
bool MAD::MP3Player<Buffer>::Rewind()
{
    Clean();
    timer_ = mad_timer_zero;

    try {

        Init();

    } catch(const Exception::SaguaException& ex) {

        return false;
    }

    decodeState_ = SEEK;
    mad_decoder_run(&decoder_, MAD_DECODER_MODE_SYNC);

    return true;
}

/// @brief Buffering file to decode.
template <class Buffer>
enum mad_flow MAD::MP3Player<Buffer>::Input(void* data, struct mad_stream* stream)
{
    struct buffer* buf = static_cast<struct buffer*>(data);

    if(buf->eof)
        return MAD_FLOW_STOP;

    if(stream->next_frame) {

        struct stat stat;
	
        if(-1 == fstat(buf->fd, &stat))
            return MAD_FLOW_BREAK;
	
        const unsigned long posn = stream->next_frame - buf->fdm;

        // If you use linux, you get warning about type 
        if(stat.st_size > buf->length) {

            if(-1 == munmap(buf->fdm, buf->length)) {

                buf->fdm = 0;
                buf->data = 0;

                return MAD_FLOW_BREAK;	
            }

            buf->length = stat.st_size;
            buf->fdm = static_cast<unsigned char*>( mmap(0, buf->length, PROT_READ, MAP_SHARED, buf->fd, 0) );

            if(buf->fdm == 0) {
                buf->data = 0;
                return MAD_FLOW_BREAK;
            }

            mad_stream_buffer(stream, buf->fdm + posn, buf->length - posn);

            return MAD_FLOW_CONTINUE;
        }

        const unsigned long left = buf->length - posn;
	
        buf->data = static_cast<unsigned char*>( malloc(left + MAD_BUFFER_GUARD) ); 
        if(buf->data == 0)
            return MAD_FLOW_BREAK;
	
        buf->eof = true;
	
        memcpy(buf->data, buf->fdm + posn, left);
        memset(buf->data + left, 0, MAD_BUFFER_GUARD);
	
        mad_stream_buffer(stream, buf->data, left + MAD_BUFFER_GUARD);
	
        return MAD_FLOW_CONTINUE;    
    }

    mad_stream_buffer(stream, buf->fdm, buf->length);
    
    return MAD_FLOW_CONTINUE;
}

/// @brief Output decoded samples.
///
/// Check audio device and open device if need. Put samples to container and
/// send to play function. Update timer and signal to MainManager if timer
/// changed in second.
template <class Buffer>
enum mad_flow MAD::MP3Player<Buffer>::Output(void* data, struct mad_header const* header, struct mad_pcm* pcm)
{
    if(!playerInterface_->CheckOpenDevice()) {

        buffer_->Close(true);
        if(!playerInterface_->OpenDevice(16, pcm->channels, pcm->samplerate))
            return MAD_FLOW_STOP;
    }
    
    if(pcm->channels == 2) {
	
        buffer_->Push(pcm->samples[0], pcm->samples[1], pcm->length);
	
    } else {

        buffer_->Push(pcm->samples[0], pcm->length);
    }

    mad_timer_add(&timer_, header->duration);
 
    return MAD_FLOW_CONTINUE;
}

/// @brief Control decoder. 
/// @todo Put some cut-off audio output while SEEKING
template <class Buffer>
enum mad_flow MAD::MP3Player<Buffer>::Header(void* data, struct mad_header const* header)
{
    if(decodeState_ == STOP)
        return MAD_FLOW_STOP;
    
    if(decodeState_ == SEEK) {

        if(timer_.seconds < seekSeconds_) { 
            mad_timer_add(&timer_, header->duration);
    
            return MAD_FLOW_IGNORE;

        } else if(timer_.seconds > seekSeconds_) {

            decodeState_ = SEEK_REWIND;
	    	
            return MAD_FLOW_STOP;

        } else {

            seekSeconds_ = 0;

            decodeState_ = PLAY;
        }
    }
	    
    return MAD_FLOW_CONTINUE;
}

/// @brief Create MP3Player object.
///
/// Used in PlayerFactory.
template <class Buffer>
std::auto_ptr<Player> CreateMP3Player(Buffer& buffer, const std::string& filename)
{
    return std::auto_ptr<Player>(new MAD::MP3Player<Buffer>(buffer, filename));
}
