/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file playlist.cc
/// @brief Class for Playlist source file.

#include <iostream>
#include <fstream>

#include <vector>
#include <list>
#include <map>

#include <cstdio>

#include <gdkmm.h>
#include <gtkmm/stock.h>

#include <libintl.h>

#include "../include/playlist.h"
#include "../include/mainwin.h"
#include "../include/tag_empty.h"
#include "../include/tag_mp3.h"
#include "../include/tag_ogg.h"
#include "../include/tag_wma.h"

#ifdef ENABLE_FAAD2
#include "../include/tag_aac.h"
#endif

#ifdef ENABLE_HTTP
#include "../include/connector.h"
#endif

#include "../include/fileinfodlg.h"
#include "../include/mainmanager.h"
#include "../include/playcontroller.h"
#include "../include/fixedallocator.h"
#include "../include/misc.h"
#include "../include/trace.h"
#include "../include/enums.h"

class PlaylistImpl::EachRow : public SmallObject
{
public:
    EachRow(const Gtk::TreeModel::iterator& iter, std::auto_ptr<Tag> tag)
        : iter_(iter), tag_(tag)
    {}

    void Destroy()
    {
        delete this;
    }

    Gtk::TreeModel::iterator iter_;
    const std::auto_ptr<Tag> tag_;

private:
    ~EachRow()
    {}

    EachRow(const EachRow&);
    EachRow operator=(const EachRow&);
};

const unsigned int PlaylistImpl::mainWndPanelHeight_(170);
const unsigned int PlaylistImpl::eachRowHeight_(25);

PlaylistImpl::PlaylistImpl(AddItem& addItem, PlayListControl& playCtl, unsigned int mainWndHeight, unsigned int titleWidth, const bool& loadlist)
    : count_(0),
      tagProcessCount_(0),
      inTagParsing_(false),
      stopTagParsing_(false),
      treeModel_(Gtk::ListStore::create(columns_)),
      tagFactory_(new TagFactory),
      listFactory_(new ListFactory),
      playCtl_(playCtl),
      tagParsingThread_(0)
{
    CalcTagProcessCount(mainWndHeight - mainWndPanelHeight_);

    set_model(treeModel_);

    SetFontSize(10);

    append_column(gettext("Number"), columns_.id_);
    append_column(gettext("Title"), columns_.title_);
    append_column(gettext("Time"), columns_.time_);

    CreatePopupMenuList(addItem);

    SetTitleColumnProperties(titleWidth);

    RegisterTagFactory();

    RegisterListType();

    if(loadlist)
        LoadLastList();

    treeModel_->set_sort_func(columns_.filename_, sigc::mem_fun(*this, &PlaylistImpl::FilenameCompare));

    signalRowToFill_.connect(sigc::mem_fun(*this, &PlaylistImpl::FillRowFromQueue));

    std::list<Gtk::TargetEntry> targets;
    targets.push_back( Gtk::TargetEntry("text/plain") );
    drag_dest_set(targets);

    ForceUpdateToppage();
}

PlaylistImpl::~PlaylistImpl()
{
    StopThread();

    SaveList();
}

/// @brief Add file to list interface function.
///
/// If file is not playable and file is not list-type file, 
/// file will be rejected. Or call AddFile(const std::string& filePath) 
/// function to add a file that not parsed information.
/// @param filePath A filepath to add.
/// @param addListFile Load listfile if filepath is list-type file.
void PlaylistImpl::AddFile(const std::string& filePath, const bool& addListFile)
{
    if(playCtl_.IsPlayable(filePath)) {

        AddFile(filePath);

    } else if(addListFile && IsListType(filePath)) {

        LoadList(filePath);

    }
}

Gtk::Widget& PlaylistImpl::GetWidget()
{
    return *this;
}

/// @brief Add files to list.
/// @param files Files to add. 
/// @param addListFile Load listfile if filepath is list-type file.
void PlaylistImpl::AddFiles(const std::vector<std::string>& files, const bool& addListFile)
{
    typedef std::vector<std::string>::const_iterator iter;

    const iter end(files.end());
    for(iter eachFile(files.begin()); eachFile != end; ++eachFile) {

        AddFile(*eachFile, addListFile);

    }
}

/// @brief Add files under directory to playlist.
/// @param directoryPath Directory to add.
void PlaylistImpl::AddDirectory(const std::string& directoryPath)
{
    Glib::Dir dir(directoryPath);

    std::string fullpath;
    const Glib::Dir::const_iterator dirEnd(dir.end());

    for(Glib::Dir::const_iterator eachFile = dir.begin(); eachFile != dirEnd; ++eachFile) {

        fullpath = directoryPath;
        fullpath += '/';
        fullpath += *eachFile;

        if(!IsDirectory(fullpath) && playCtl_.IsPlayable(fullpath))
            AddFile(fullpath);
    }
}

/// @brief Add files under directory to playlist recursively.
/// @param path Directory to add.
void PlaylistImpl::AddDirectoryRecursive(const std::string& path)
{
    Glib::Dir dir(path);

    std::string fullpath;
    // Very ugly repeat codes...
    for(Glib::Dir::const_iterator eachFile = dir.begin(); eachFile != dir.end(); ++eachFile) {

        fullpath = path;
        fullpath += '/';
        fullpath += *eachFile;

        if(IsDirectory(fullpath)) {
            // Repeat again or recursive call?
            if(IsExecutable(fullpath) && IsReadable(fullpath)) {
                Glib::Dir dir2(fullpath);
		    
                std::string fullpath2;
                for(Glib::Dir::const_iterator eachFile2 = dir2.begin(); eachFile2 != dir2.end(); ++eachFile2) {

                    fullpath2 = fullpath;
                    fullpath2 += '/';
                    fullpath2 += *eachFile2;

                    if(IsDirectory(fullpath2)) {
                        if(IsExecutable(fullpath2) && IsReadable(fullpath2))
                            // If have more depth, call again anyway.
                            AddDirectoryRecursive(fullpath2);
                    } else if(playCtl_.IsPlayable(fullpath2))
                        AddFile(fullpath2);
                }
            }
        } else if(playCtl_.IsPlayable(fullpath))
            AddFile(fullpath);
    }
}

#ifdef ENABLE_HTTP

/// @brief Add url to playlist interface function.
///
/// If url is INET file, just add it.
/// If url is some INET list file, get it and add it's contents.
/// Or add a url as stream.
/// @param Url to add.
void PlaylistImpl::AddUrl(const std::string& destUrl)
{
    if(IsListType(destUrl)) {

        const std::string tmpFile(Glib::get_tmp_dir() + "/sagua_tmplist-" + Glib::get_user_name() + "." + GetExt(destUrl));

        if(!Unlink(tmpFile))
            return;

        GetListFromUrl(destUrl, tmpFile);

        LoadList(tmpFile);

        Unlink(tmpFile);

    } else {

        const Glib::ustring convertedDestURL(StringToUTF8::Instance().Convert(destUrl));

        AddFile(convertedDestURL, convertedDestURL, "--:--", 1);

    }
}

void PlaylistImpl::GetListFromUrl(const std::string& destUrl,
                              const std::string& listFile)
{
    // reduce these duplicated codes.
    bool redirect = false;
    unsigned int retry = 0;
    std::auto_ptr<Address> destAddress;
    std::string url(destUrl);

    do {

        destAddress = CreateAddress(url, signalConvertUrlEncoding_());

        if(0 == destAddress.get())
            return;

        ACE_INET_Addr srvr(destAddress->GetPortNumber(),
                           destAddress->GetRawHostName());

        ACE_Connector<FileGet, ACE_SOCK_CONNECTOR> connector;

        FileGet fileGet;

        HttpRequestStrGenerator generator(*destAddress);

        fileGet.SetGenerator(&generator);
        fileGet.SetOutFile(listFile.c_str());
        FileGet* currentFileGet = &fileGet;
        
        if(connector.connect(currentFileGet, srvr) != -1) {

            ACE_Reactor::instance()->reset_reactor_event_loop();
            ACE_Reactor::instance()->run_reactor_event_loop();

        }

        if(redirect = fileGet.Redirect())
            url = fileGet.RedirectURL();

    } while(redirect && retry++ < 3);
}

#endif

void PlaylistImpl::RemoveAll()
{
    OnMenuRemoveAll();
}

std::string PlaylistImpl::DoGetCurrent(Sagua::ListState& listState)
{
    if(IsListEmpty())
        return EmptyNullFilename(listState);

    if(!pos_)
        pos_ = treeModel_->children().begin();

    return Get(pos_, listState);
}

std::string PlaylistImpl::DoGetNext(Sagua::ListState& listState)
{
    if(IsListEmpty())
        return EmptyNullFilename(listState);

    if(!pos_)
        return GetFirst(listState);

    if(IsPositionLastRow())
        return LastNullFilename(listState);

    return Get(++pos_, listState);
}

std::string PlaylistImpl::DoGetPrev(Sagua::ListState& listState)
{
    if(IsListEmpty())
        return EmptyNullFilename(listState);

    if(!pos_)
        return GetFirst(listState);

    if(IsPositionFirstRow())
        return FirstNullFilename(listState);

    return Get(--pos_, listState);
}

std::string PlaylistImpl::DoGetFirst(Sagua::ListState& listState)
{
    if(IsListEmpty())
        return EmptyNullFilename(listState);

    pos_ = treeModel_->children().begin();

    return Get(pos_, listState);
}

std::string PlaylistImpl::DoGetLast(Sagua::ListState& listState)
{
    if(IsListEmpty())
        return EmptyNullFilename(listState);

    pos_ = --(treeModel_->children().end());

    return Get(pos_, listState);
}

std::string PlaylistImpl::DoGetRandom(Sagua::ListState& listState)
{
    if(count_ == 0)
        return EmptyNullFilename(listState);

    MoveNthPosition(pos_, GetRandomNumber(pos_));
    
    return Get(pos_, listState);
}

/// @brief Force tag parsing from top page.
void PlaylistImpl::ForceUpdateToppage()
{
    const Gtk::TreeModel::iterator iter(treeModel_->children().begin());

    if(iter) {
        StopThread();
        inTagParsing_ = true;
        tagParsingThread_ = Glib::Thread::create( sigc::bind( sigc::mem_fun(*this, &PlaylistImpl::GetInfoFromTag), iter), true );
    }
}

/// @brief Force tag parsing from current view page.
void PlaylistImpl::ForceUpdatePage()
{
    while(Glib::MainContext::get_default()->iteration(false));

    OnVScrolled();
}

Glib::ustring PlaylistImpl::CurrentItemTitle()
{
    return (*pos_)[columns_.title_];
}

Glib::ustring PlaylistImpl::CurrentItemTime()
{
    return (*pos_)[columns_.time_];
}

unsigned int PlaylistImpl::CurrentItemTimeSeconds()
{
    return (*pos_)[columns_.timeSeconds_];
}

/// @brief Show current selected item's information.
void PlaylistImpl::ShowCurrentFileInfoDlg() const
{
    if(pos_)
        ShowFileInfoDlg( (*pos_)[columns_.filename_] );
}

void PlaylistImpl::SetTitleWidth(unsigned int titleWidth)
{
    get_column(1)->set_fixed_width(titleWidth);
}

/// @brief Callback function used when button is press.
///
/// If pressed button is 3 button, pop up menu.
bool PlaylistImpl::on_button_press_event(GdkEventButton* ev)
{
    const bool rtn(TreeView::on_button_press_event(ev));

    if( (ev->type == GDK_BUTTON_PRESS) && (ev->button == 3) ) {
        menuPopup_.popup(ev->button, ev->time);
    }

    return rtn;
}

/// @brief Callback function used when row is activated(double clicked or else ...)
void PlaylistImpl::on_row_activated(const Gtk::TreeModel::Path& path, Gtk::TreeViewColumn* column)
{
    Gtk::TreeView::on_row_activated(path, column);

    const Gtk::TreeModel::iterator iter(treeModel_->get_iter(path));

    if(iter) {
        pos_ = iter;

        playCtl_.PlayFile(Get(pos_));
    }
}

/// @brief Callback function used when scroll adjustments is set.
///
/// This function called when TreeView regist hadj, vadj and unregist them.
/// OnVScrolled is connected vadj's signal_value_changed().
void PlaylistImpl::on_set_scroll_adjustments(Gtk::Adjustment* hadj, Gtk::Adjustment* vadj)
{
    Gtk::TreeView::on_set_scroll_adjustments(hadj, vadj);

    if(vadj) {
        vadj->signal_value_changed().connect( sigc::mem_fun(*this, &PlaylistImpl::OnVScrolled) );
    }

}

/// @brief Callback function used when size allocate.
///
/// This function called when widget is added to other container. 
/// tagProcessCount_ is calculated in here.
void PlaylistImpl::on_size_allocate(Gtk::Allocation& allocation)
{
    Gtk::TreeView::on_size_allocate(allocation);

    CalcTagProcessCount(allocation.get_height());
}

/// @brief Fill row from rowQueue.
///
/// This function must be called in thread.
void PlaylistImpl::FillRowFromQueue()
{
    EachRow* processRow;

    {
        Glib::Mutex::Lock lock(tagParsingMutex_);

        if(rowQueue_.empty())
            return;

        processRow = rowQueue_.front();
        rowQueue_.pop();
    }

    if(processRow->iter_)
        FillRow(processRow->iter_, processRow->tag_->GetTitle(), processRow->tag_->GetTotalTime(), processRow->tag_->GetTotalTimeSeconds());

    processRow->Destroy();
}

/// @brief Parse tag loop.
/// @param iter Start iterator
void PlaylistImpl::GetInfoFromTag(Gtk::TreeModel::iterator& iter)
{
    unsigned int i = 0;

    while(!stopTagParsing_) {

        if(!iter || i > tagProcessCount_)
            break;

        {
            Glib::Mutex::Lock lock(tagParsingMutex_);

            if(!(*iter)[columns_.tagParsed_]) {

                const std::string filename((*iter)[columns_.filename_]);

                if(IsReadable(filename)) {

                    std::auto_ptr<Tag> tagParser(tagFactory_->Create(filename));

                    if(0 != tagParser.get()) {

                        rowQueue_.push(new EachRow(iter, tagParser));

                        signalRowToFill_();
                    }

                } else {

                    std::auto_ptr<Tag> tagParser(CreateEmptyTag(filename));

                    if(0 != tagParser.get()) {

                        rowQueue_.push(new EachRow(iter, tagParser));

                        signalRowToFill_();
                    }
                }
            }
        }

        ++iter;
        ++i;
    }
}

/// @brief Show information dialog.
void PlaylistImpl::ShowFileInfoDlg(const std::string& filename) const
{
    if(IsReadable(filename)) {
        std::auto_ptr<Tag> tagParser(tagFactory_->Create(filename));

        if(0 != tagParser.get()) {

            tagParser->ProcessAll();

            FileInfoDlg fileInfoDlg(*tagParser);
            fileInfoDlg.Run();
        }
    }
}

/// @brief Callback function used when drag data received.
///
/// When drag data received, add them to list.
void PlaylistImpl::on_drag_data_received(const Glib::RefPtr<Gdk::DragContext>& context, int x, int y, const Gtk::SelectionData& selectionData, guint info, guint time)
{
    if((selectionData.get_length() >=0 ) && (selectionData.get_format() == 8)) {

        AddDragedFiles(selectionData.get_data_as_string());

        ForceUpdateToppage();

    }

    context->drag_finish(false, false, time);
}

/// @brief Callback function used when remove menu is clicked.
void PlaylistImpl::OnMenuRemove()
{
    StopThread();

    if(Glib::RefPtr<Gtk::TreeView::Selection> selection = get_selection()) {
        if(Gtk::TreeModel::iterator iter = selection->get_selected()) {
            if(pos_ && pos_ == iter) {

                playCtl_.ItemRemoved();

                if(IsPositionLastRow())
                    pos_ = treeModel_->children().begin();
                else
                    ++pos_;

            }

            unsigned int count = (*iter)[columns_.id_];

            iter = treeModel_->erase(iter);
            MoveCursorAndScroll(iter);
            count_--;
            
            // Renewal id number. Do you have more effiecient idea?
            iter = treeModel_->children().begin();
            const Gtk::TreeModel::iterator end(treeModel_->children().end());
            while(iter != end) {
                if((*iter)[columns_.id_] > count) {
                    (*iter)[columns_.id_] = count;
                    ++count;
                }

                ++iter;
            }
        }
    }
}

/// @brief Callback function used when remove all menu is clicked.
void PlaylistImpl::OnMenuRemoveAll()
{
    StopThread();

    playCtl_.ItemRemoved();

    ClearRowQueue();

    treeModel_->clear();

    count_ = 0;
    pos_ = treeModel_->children().begin();
}

/// @brief Callback function used when sort filename menu is clicked.
void PlaylistImpl::OnMenuSortFilename()
{
    treeModel_->set_sort_column(columns_.filename_, Gtk::SORT_ASCENDING);

    ForceUpdatePage();

    MoveCursorAndScroll(pos_);
}

/// @brief Callback function used when information menu is clicked.
void PlaylistImpl::OnMenuInfo()
{
    if(Glib::RefPtr<Gtk::TreeView::Selection> selection = get_selection()) {
        if(Gtk::TreeModel::iterator iter = selection->get_selected()) {

            ShowFileInfoDlg( (*iter)[columns_.filename_] );

        }
    }

}

/// @brief Callback function used when scrolled.
void PlaylistImpl::OnVScrolled()
{
    if(IsListEmpty())
        return;	

    StopThread();

    inTagParsing_ = true;
    tagParsingThread_ = Glib::Thread::create( sigc::bind( sigc::mem_fun(*this, &PlaylistImpl::GetInfoFromTag), GetTopPositionIterator()), true );
}

/// @brief Load list from indentification string file.
void PlaylistImpl::LoadLastList()
{
    DEBUGTRACE("PlaylistImpl::LoadLastList");

    LoadList(ProgramFile::Instance().GetMainList());
}

/// @brief Load list from filename.
void PlaylistImpl::LoadList(const std::string& filename)
{
    DEBUGTRACE("PlaylistImpl::LoadList");

    std::auto_ptr<ListFile> listFileProcessor(listFactory_->Create(filename));

    if(0 != listFileProcessor.get())
        listFileProcessor->Read(filename, *this);
}

/// @brief Save current list to identification string file.
void PlaylistImpl::SaveList()
{
    DEBUGTRACE("PlaylistImpl::SaveList");

    if(!ProgramFile::Instance().CheckDir())
        return;

    M3UList listFile;
    listFile.Write(*this, ProgramFile::Instance().GetMainList());
}

/// @brief Stop thread.
void PlaylistImpl::StopThread()
{
    if(inTagParsing_) {
        stopTagParsing_ = true;
        tagParsingThread_->join();
        inTagParsing_ = false;	
        stopTagParsing_ = false;
    }
}

/// @brief Create popup menu list.
void PlaylistImpl::CreatePopupMenuList(AddItem& addItem)
{
    Gtk::Menu_Helpers::MenuList& popupMenuList = menuPopup_.items();
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Add files"), sigc::mem_fun(addItem, &AddItem::FileAdd) ));
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("Add a _directory"), sigc::mem_fun(addItem, &AddItem::DirAdd) ));
#ifdef ENABLE_HTTP
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("Add a URL"), sigc::mem_fun(addItem, &AddItem::UrlAdd) ));
#endif
    popupMenuList.push_back( Gtk::Menu_Helpers::SeparatorElem() );
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Remove"), sigc::mem_fun(*this, &PlaylistImpl::OnMenuRemove) ));
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("Re_move all"), sigc::mem_fun(*this, &PlaylistImpl::OnMenuRemoveAll) ));
    popupMenuList.push_back( Gtk::Menu_Helpers::SeparatorElem() );
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Sort filename"), sigc::mem_fun(*this, &PlaylistImpl::OnMenuSortFilename) ));
    popupMenuList.push_back( Gtk::Menu_Helpers::SeparatorElem() );
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Information"), sigc::mem_fun(*this, &PlaylistImpl::OnMenuInfo) ));
}

/// @brief Register creator function to TagFactory.
void PlaylistImpl::RegisterTagFactory()
{
    tagFactory_->Register("empty", &CreateEmptyTag);
    tagFactory_->Register("mp1", &CreateId3Tag);
    tagFactory_->Register("mp2", &CreateId3Tag);
    tagFactory_->Register("mp3", &CreateId3Tag);
    tagFactory_->Register("ogg", &CreateOGGTag);
    tagFactory_->Register("wma", &CreateWMATag);

#ifdef ENABLE_FAAD2
    tagFactory_->Register("mp4", &CreateAACTag);
    tagFactory_->Register("m4a", &CreateAACTag);
    tagFactory_->Register("aac", &CreateAACTag);
#endif
}

void PlaylistImpl::RegisterListType()
{
    listFactory_->Register("m3u", &CreateM3UList);
    listFactory_->Register("pls", &CreatePLSList);
}

bool PlaylistImpl::IsListType(const std::string& filePath)
{
    return listFactory_->Exist(filePath);
}

/// @brief Add files from DND.
void PlaylistImpl::AddDragedFiles(const std::string& dragedString)
{
    std::string::const_iterator itemBegin(dragedString.begin());
    std::string::const_iterator itemEnd;
    const std::string::const_iterator end(dragedString.end());
    std::string eachItem;

    for(;;) {

        itemBegin = find_if(itemBegin, end, std::not1(IsCharSpace()));

        if(itemBegin == end)
            break;

        itemEnd = find_if(itemBegin, end, IsCharSpace());

        if(itemBegin != itemEnd) {

            eachItem = std::string(itemBegin, itemEnd);

            if(HasFileProtocolString(eachItem)) {

                try {
                    eachItem = Glib::filename_from_uri(eachItem);

                    if(IsDirectory(eachItem))
                        AddDirectoryRecursive(eachItem);
                    else
                        AddFile(eachItem, true);

                } catch(Glib::ConvertError& err) {

                    // Got convert error!
                }
            }

#ifdef ENABLE_HTTP
            if(HasHttpProtocolString(eachItem))
                AddUrl(eachItem);
#endif
        }

        itemBegin = itemEnd;
    }
}

int PlaylistImpl::FilenameCompare(const Gtk::TreeModel::iterator& a, const Gtk::TreeModel::iterator& b)
{
    const std::string strA((*a)[columns_.filename_]);
    const std::string strB((*b)[columns_.filename_]);

    const std::string aPath(Glib::path_get_dirname(strA));
    const std::string bPath(Glib::path_get_dirname(strB));

    // If files in same directory, sort them
    // or sort for directory.	
    const int rtn = std::strncmp(aPath.c_str(), bPath.c_str(), aPath.size());
    if(rtn == 0)
        return std::strncmp(strA.c_str(), strB.c_str(), strA.size());
    else
        return rtn;
}

void PlaylistImpl::SetFontSize(unsigned int size)
{
    Pango::FontDescription font = get_style()->get_font();
    font.set_size( Pango::SCALE * size );
    Glib::RefPtr<Gtk::RcStyle> style = Gtk::RcStyle::create();
    style->set_font(font);
    modify_style(style);
}

void PlaylistImpl::TagParsing(Gtk::TreeModel::iterator& iter)
{
    Glib::Mutex::Lock lock(tagParsingMutex_);

    if( !(*iter)[columns_.tagParsed_] ) { 

        const std::string filename((*iter)[columns_.filename_]);

        if(IsReadable(filename)) {

            std::auto_ptr<Tag> tagParser(tagFactory_->Create(filename));

            if(0 != tagParser.get())
                FillRow(iter, tagParser->GetTitle(), tagParser->GetTotalTime(), tagParser->GetTotalTimeSeconds()); 
        } else {

            std::auto_ptr<Tag> tagParser(CreateEmptyTag(filename));

            if(0 != tagParser.get())
                FillRow(iter, tagParser->GetTitle(), tagParser->GetTotalTime(), tagParser->GetTotalTimeSeconds());
        }
    }
}

std::string PlaylistImpl::Get(Gtk::TreeModel::iterator& iter)
{
    TagParsing(iter);

    MoveCursorAndScroll(iter);

    return std::string( (*iter)[columns_.filename_] );
}

/// @brief Get filename from iter.
///
/// This function don't check validation of iter.
std::string PlaylistImpl::Get(Gtk::TreeModel::iterator& iter, Sagua::ListState& listState)
{
    listState = Sagua::LIST_NORMAL;

    return Get(iter);
}

unsigned int PlaylistImpl::GetRandomNumber(Gtk::TreeModel::iterator& iter)
{
    // Prevent undefined behavior
    if(count_ == 1)
        return 1;

    Glib::Rand rand;

    if(!iter)
        iter = treeModel_->children().begin();

    // I don't want get same element.
    const unsigned int previousNumber = (*iter)[columns_.id_] - 1;
    unsigned int randomNumber = 0;
    do {

        randomNumber = rand.get_int_range(0, count_);

    } while(randomNumber == previousNumber);

    return randomNumber;
}

void PlaylistImpl::MoveNthPosition(Gtk::TreeModel::iterator& iter, const unsigned int& number)
{
    iter = treeModel_->children().begin();

    for(unsigned int i = 0; i != number; ++i)
        ++iter;
}

/// @brief Get top position of treeview iterator.
Gtk::TreeModel::iterator PlaylistImpl::GetTopPositionIterator()
{
    // useless variables
    Gtk::TreeViewColumn* eachColumn;
    int cellX = 0;
    int cellY = 0;

    Gtk::TreePath eachPath;
    get_path_at_pos(0, 0, eachPath, eachColumn, cellX, cellY);

    return treeModel_->get_iter(eachPath);
}

/// @brief Move cursor to coresspond iterator and scroll to that.
void PlaylistImpl::MoveCursorAndScroll(const Gtk::TreeModel::iterator& iter)
{
    Gtk::TreeModel::Path path = iter ? Gtk::TreeModel::Path(iter) :
        Gtk::TreeModel::Path("0");

    set_cursor(path);
    scroll_to_row(path, 0.5);
}

/// @brief Destroy all row queue data and empty it.
void PlaylistImpl::ClearRowQueue()
{
    Glib::Mutex::Lock lock(tagParsingMutex_);

    while(!rowQueue_.empty()) {

        EachRow* row(rowQueue_.front());

        rowQueue_.pop();

        row->Destroy();
    }
}

void M3UList::Read(const std::string& filename, PlaylistImpl& output)
{
    std::ifstream listFile(filename.c_str());
    if(listFile) {

        std::string eachLine;
        eachLine.reserve(255);
        const std::string dirname(Glib::path_get_dirname(filename) + '/');
        std::string tmp;
        tmp.reserve(255);

        while(getline(listFile, eachLine)) {

            if(IsComment(eachLine))
                continue;
            else {
                TrimLastSpaceCharacter(eachLine);
                // If line is not absolute path, assume that file in list file's dir.

                if(Glib::path_is_absolute(eachLine))
                    AddFile(eachLine, output);
#ifdef ENABLE_HTTP
                else if(HasHttpProtocolString(eachLine))
                    AddUrl(eachLine, output);
#endif
                else {
                    tmp = dirname;
                    tmp += eachLine;
                    AddFile(tmp, output);
                }
            }
        }
    }
}

void M3UList::Write(const PlaylistImpl& input, const std::string& filename)
{
    FILE* listFile = std::fopen(filename.c_str(), "w");
    if(listFile) {

        // m3u format... (is m3u a media queue format? anyway)
        //	#EXTM3U
        //	#EXTINF:song_length_in_seconds,song_title
        //	somefilepath.mp3

        std::fprintf(listFile, "#EXTM3U\n");

#warning "hey, expection for not parsed file?"
        Gtk::TreeModel::iterator iter(GetFirstIter(input));

        unsigned int seconds = 0;
        Glib::ustring title;
        std::string filename;

        while(iter) {

            GetRawInformation(input, iter, seconds, title, filename);

            std::fprintf(listFile,
                         "#EXTINF:%d,%s\n%s\n",
                         seconds,
                         title.c_str(),
                         filename.c_str());
            ++iter;
        }

        std::fclose(listFile);
    }
}

void PLSList::Read(const std::string& filename, PlaylistImpl& output)
{
    std::ifstream listFile(filename.c_str());
    if(listFile) {

        std::string eachLine;
        std::string::iterator iter;
        const std::string dirname(Glib::path_get_dirname(filename) + '/');
        std::string tmp;

        while(getline(listFile, eachLine)) {

            if(std::strncmp(eachLine.c_str(), "File", 4) == 0) {

                iter = find_if(eachLine.begin(), eachLine.end(), IsCharEqual());
                if(iter != eachLine.end()) {

                    eachLine = std::string(++iter, eachLine.end());
                    TrimLastSpaceCharacter(eachLine);

                    if(Glib::path_is_absolute(eachLine))
                        AddFile(eachLine, output);
#ifdef ENABLE_HTTP
                    else if(HasHttpProtocolString(eachLine))
                        AddUrl(eachLine, output);
#endif
                    else {

                        tmp = dirname;
                        tmp += eachLine;
                        AddFile(tmp, output);
                    }
                }
            }
        }
    }
}

void PLSList::Write(const PlaylistImpl& input, const std::string& filename)
{
    std::ofstream listFile(filename.c_str());
    if(listFile) {

        // pls format
        //
        // [playlist]
        // File1=some filename
        // Title1=some title
        // Length1=some length in seconds
        //			(negative number will be ignored, some stream)
        // File2=blah blah
        // Title2=blah blah
        // Length2=blah blah
        // 		...
        // NumberOfEntries=totle number of entry
        // Version=2
        //
        // All string case sensitive.
        listFile << "[playlist]" << std::endl;

#warning "hey, expection for not parsed file?"
        Gtk::TreeModel::iterator iter(GetFirstIter(input));

        unsigned int count = 0;
        unsigned int seconds = 0;
        Glib::ustring title;
        std::string filename;

        while(iter) {

            ++count;

            GetRawInformation(input, iter, seconds, title, filename);
            listFile << "File" << count << "=" << filename << std::endl;
            listFile << "Title" << count << "=" << title << std::endl;
            listFile << "Length" << count << "=" << seconds << std::endl;

            ++iter;

        }

        listFile << "NumberOfEntries=" << count << std::endl;
        listFile << "Version=2" << std::endl;
    }
}

std::auto_ptr<ListFile> CreateM3UList()
{
    return std::auto_ptr<ListFile>(new M3UList);
}

std::auto_ptr<ListFile> CreatePLSList()
{
    return std::auto_ptr<ListFile>(new PLSList);
}
