/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file preference.cc
/// @brief Class for user preference setting source file.

#include <cstdlib>
#include <iostream>

#include "../include/preference.h"
#include "../include/misc.h"

SaguaPreference::SaguaPreference()
    : zeroDouble_(0.0),
      successLoadConfig_(Load())
{
    effectUse_= GetBool("Effect");

    StringToUTF8::Instance().SetAssumeCharset(GetString("GeneralAssumeCharset"));
}

void SaguaPreference::Save(std::ostream& os)
{
    SaveElem(stringElem_, os);
    SaveElem(doubleElem_, os);
    SaveElem(floatElem_, os);
    SaveElem(boolElem_, os);
}

const std::string& SaguaPreference::DoGetString(const char* key) const
{
    std::map<std::string, std::string>::const_iterator iter(stringElem_.find(key));

    if(iter != stringElem_.end())
        return iter->second;
    else
        return nullString_;
}

const double& SaguaPreference::DoGetDouble(const char* key) const
{
    std::map<std::string, double>::const_iterator iter(doubleElem_.find(key));

    if(iter != doubleElem_.end())
        return iter->second;
    else
        return zeroDouble_;
}

float SaguaPreference::DoGetFloat(const char* key) const
{
    std::map<std::string, float>::const_iterator iter(floatElem_.find(key));

    if(iter != floatElem_.end())
        return iter->second;
    else
        return 0.0f;
}

bool SaguaPreference::DoGetBool(const char* key) const
{
    std::map<std::string, bool>::const_iterator iter(boolElem_.find(key));

    if(iter != boolElem_.end())
        return iter->second;
    else
        return false;
}

void SaguaPreference::GetAudioDriverOption(std::map<std::string, std::string>& dest)
{
    const std::string& driver(GetString("AudioDriver"));

    if(driver == "esd" || driver == "default")
        return;

    if(driver == "alsa") {

        dest["card"] = GetString("AudioAlsa5Card");
        dest["dev"] = GetString("AudioAlsa5Dev");

    } else if(driver == "alsa09") {

        dest["dev"] = GetString("AudioAlsa9Dev");
    
    } else if(driver == "oss") {

        dest["dsp"] = GetString("AudioOssDsp");

    } else if(driver == "sun") {

        dest["dev"] = GetString("AudioSunDev");

    } else if(driver == "aixs") {

        dest["dev"] = GetString("AudioAixsDev");

    }
}

void SaguaPreference::Set(const char* key, const std::string& value)
{
    stringElem_[key] = value;
}

void SaguaPreference::Set(const char* key, const double& value)
{
    doubleElem_[key] = value;
}

void SaguaPreference::Set(const char* key, float value)
{
    floatElem_[key] = value;
}

void SaguaPreference::DoSet(const char* key, bool value)
{
    boolElem_[key] = value;
}

bool SaguaPreference::Load()
{
    std::map<std::string, std::string> configMap;

    if(ConfigfileToMap(ProgramFile::Instance().GetConfig(), configMap, false, false)) {

        Insert(configMap, "GeneralTrayIconAction", stringElem_, "fileinfo");
        Insert(configMap, "GeneralAssumeCharset", stringElem_, "eucKR");
        Insert(configMap, "GeneralCompactBgPixmap", stringElem_, "");
        Insert(configMap, "MixerDev", stringElem_, "/dev/mixer");
        Insert(configMap, "GeneralConvertUrlToCurrentEncoding", boolElem_, false);
        Insert(configMap, "AudioDriver", stringElem_, "default");
        Insert(configMap, "AudioAlsa5Card", stringElem_, "0");
        Insert(configMap, "AudioAlsa5Dev", stringElem_, "0");
        Insert(configMap, "AudioAlsa9Dev", stringElem_, "default");
        Insert(configMap, "AudioOssDsp", stringElem_, "/dev/dsp");
        Insert(configMap, "AudioSunDev", stringElem_, "/dev/audio");
        Insert(configMap, "AudioAixsDev", stringElem_, "/dev/baud0/1");
        Insert(configMap, "EfxStereoEnhanceAmount", doubleElem_, 1.0);
        Insert(configMap, "EfxRoomAmount", doubleElem_, 0.4);
        Insert(configMap, "EfxRoom2Amount", doubleElem_, 0.1);
        Insert(configMap, "EfxHarmonyAmount", doubleElem_, 0.4);
        Insert(configMap, "EfxRhythmeAmount", doubleElem_, 1.0);
        Insert(configMap, "EfxDistortionAmount", doubleElem_, 7.0);
        Insert(configMap, "Effect", boolElem_, false);
        Insert(configMap, "EfxBassBoost", boolElem_, false);
        Insert(configMap, "EfxBassBoost2", boolElem_, false);
        Insert(configMap, "EfxStereoEnhance", boolElem_, false);
        Insert(configMap, "EfxTrebleBoost", boolElem_, false);
        Insert(configMap, "EfxRoom", boolElem_, false);
        Insert(configMap, "EfxRoom2", boolElem_, false);
        Insert(configMap, "EfxHarmony", boolElem_, false);
        Insert(configMap, "EfxRhythme", boolElem_, false);
        Insert(configMap, "EfxDistortion", boolElem_, false);
        Insert(configMap, "EfxGraphiceq", boolElem_, false);
        Insert(configMap, "EfxGraphicEQGain50", floatElem_, 0.0f);
        Insert(configMap, "EfxGraphicEQGain100", floatElem_, 0.0f);
        Insert(configMap, "EfxGraphicEQGain200", floatElem_, 0.0f);
        Insert(configMap, "EfxGraphicEQGain400", floatElem_, 0.0f);
        Insert(configMap, "EfxGraphicEQGain800", floatElem_, 0.0f);
        Insert(configMap, "EfxGraphicEQGain1.5k", floatElem_, 0.0f);
        Insert(configMap, "EfxGraphicEQGain3k", floatElem_, 0.0f);
        Insert(configMap, "EfxGraphicEQGain5k", floatElem_, 0.0f);
        Insert(configMap, "EfxGraphicEQGain7k", floatElem_, 0.0f);
        Insert(configMap, "EfxGraphicEQGain10k", floatElem_, 0.0f);
        Insert(configMap, "EfxGraphicEQGain15k", floatElem_, 0.0f);

        return true;

    } else
        return false;   
}

template <typename T>
void SaguaPreference::SaveElem(T& orig, std::ostream& os)
{
    typedef typename T::const_iterator iterator;

    iterator iter(orig.begin());
    const iterator iterEnd(orig.end());

    while(iter != iterEnd) {

        os << iter->first << " = " << iter->second << std::endl;

        ++iter;
    }
}

void SaguaPreference::Insert(std::map<std::string, std::string>& orig, const char* key, std::map<std::string, std::string>& dest, const char* initialValue)
{
    std::map<std::string, std::string>::const_iterator iter(orig.find(key));
    if(iter != orig.end())
        dest[key] = iter->second;
    else
        dest[key] = initialValue;
}

void SaguaPreference::Insert(std::map<std::string, std::string>& orig, const char* key, std::map<std::string, double>& dest, double initialValue)
{
    std::map<std::string, std::string>::const_iterator iter(orig.find(key));
    if(iter != orig.end())
        dest[key] = strtod((iter->second).c_str(), 0);
    else
        dest[key] = initialValue;
}

void SaguaPreference::Insert(std::map<std::string, std::string>& orig, const char* key, std::map<std::string, float>& dest, float initialValue)
{
    std::map<std::string, std::string>::const_iterator iter(orig.find(key));
    if(iter != orig.end())
        dest[key] = strtof((iter->second).c_str(), 0);
    else
        dest[key] = initialValue;
}

void SaguaPreference::Insert(std::map<std::string, std::string>& orig, const char* key, std::map<std::string, bool>& dest, bool initialValue)
{
    dest[key] = initialValue;

    std::map<std::string, std::string>::const_iterator iter(orig.find(key));
    if(iter != orig.end()) {

        if((iter->second).size() == 1) {

            switch(strtol((iter->second).c_str(), 0, 10)) {
            case 1:
                dest[key] = true;
                break;

            case 0:
                dest[key] = false;
                break;
		    
            default:
                break;	         
            }
        }
    }
}
