/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file tag_mp3.cc
/// @brief Class for process mp3 tag source file.

#include <cstdlib>
#include <cerrno>

#include <strings.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>

#include "../include/tag_mp3.h"
#include "../include/misc.h"
#include "../include/scopeguard.h"
#include "../include/trace.h"

MP3Tag::MP3Tag(const std::string& filename)
    : id3Tag_(filename.c_str())
{
    DEBUGTRACE("MP3Tag::MP3Tag");

    ProcessTextFieldUTF8(title_, ID3FID_TITLE);

    if(StringIsSpace(title_))
        title_ = StringToUTF8::Instance().Convert(Glib::path_get_basename(filename));

    try {

        MP3Time mp3Time(filename.c_str());

        totalTimeSeconds_ = mp3Time.GetTime();

        totalTime_ = ConvertSecondToTimeString(totalTimeSeconds_);

    } catch(const Exception::SaguaException& ex) {

        DEBUGTRACE_MSG(ex.what());
    }
}

void MP3Tag::ConvertGenreNumber()
{
    const Glib::ustring::size_type end(genre_.find(')'));

    if(Glib::ustring::npos != end) {

        genre_ = Glib::ustring(genre_, 1, end);
        const long genreNumber(std::strtol(genre_.c_str(), 0, 10));

        if(!errno && genreNumber >= 0 && genreNumber < ID3_NR_OF_V1_GENRES)
            genre_ = ID3_v1_genre_description[genreNumber];
        else
            genre_ = "Unknown";
    }
}

void MP3Tag::ProcessTextField(Glib::ustring& dest, const ID3_FrameID& frameID)
{
    ID3_Frame* id3Frame(id3Tag_.Find(frameID));
    if(0 != id3Frame) {

        ID3_Field* id3Field(id3Frame->GetField(ID3FN_TEXT));
        if(0 != id3Field) {
     
            char tmp[4096];       
            if(id3Field->Get(tmp, 4096) > 0)
                dest = tmp;
        }
    }
}

void MP3Tag::ProcessTextFieldUTF8(Glib::ustring& dest, const ID3_FrameID& frameID)
{
    ID3_Frame* id3Frame(id3Tag_.Find(frameID));
    if(0 != id3Frame) {

        ID3_Field* id3Field(id3Frame->GetField(ID3FN_TEXT));
        if(0 != id3Field) {

            char tmp[4096];
            if(id3Field->Get(tmp, 4096) > 0) {

                dest = tmp;

                if(id3Field->GetEncoding() != ID3TE_UTF8)
                    dest = StringToUTF8::Instance().Convert(dest);
            }
        }
    }
}

void MP3Tag::ProcessYear()
{
    ProcessTextField(year_, ID3FID_YEAR);
}

void MP3Tag::ProcessTrack()
{
    ProcessTextField(track_, ID3FID_TRACKNUM);
}

void MP3Tag::ProcessAlbum()
{
    ProcessTextFieldUTF8(album_, ID3FID_ALBUM);
}

void MP3Tag::ProcessGenre()
{
    ProcessTextField(genre_, ID3FID_CONTENTTYPE);
    
    if(!genre_.empty() && genre_[0] == '(')
        ConvertGenreNumber();
}

void MP3Tag::ProcessArtist()
{
    ProcessTextFieldUTF8(artist_, ID3FID_LEADARTIST);
}

void MP3Tag::ProcessComment()
{
    ProcessTextFieldUTF8(comment_, ID3FID_COMMENT);
}

/// @brief Create MP3Tag object.
///
/// Used in TagFactory.
std::auto_ptr<Tag> CreateId3Tag(const std::string& filename)
{
    return std::auto_ptr<Tag>(new MP3Tag(filename));
}

bool MP3Time::GetXingFrames(struct mad_bitptr* ptr, unsigned int bitlen, unsigned long& xingFrames)
{
    if(!HasXingHeader(ptr, bitlen))
        return false;

    long xingFlags = mad_bit_read(ptr, 32);
    bitlen -= 64;

    if(HasXingFramesFlag(xingFlags)) {

        if(bitlen < 32) {

            return false;
        }

        xingFrames = mad_bit_read(ptr, 32);

        return true;
    }

    return false;
}

bool MP3Time::DetectXingHeader(unsigned long& xingFrames)
{
    // You need to use mad_frame_decode to find ancillary data
    // that xing header in.
    frame_.header = header_;

    if(mad_frame_decode(&frame_, &stream_) != -1 &&
       GetXingFrames(&stream_.anc_ptr, stream_.anc_bitlen, xingFrames))

            return true;

    return false;
}

void MP3Time::Calc()
{
    DEBUGTRACE("MP3Time::Calc");

    unsigned long xingFrames = 0;
    FileType fileType = VBR;

    DetectFileType(fileType, xingFrames);

    switch(fileType) {
    case CBR:
        DEBUGTRACE_MSG("time calcurated in CBR");

        CalcCBR();

        break;
            
    case VBR:
        DEBUGTRACE_MSG("time calcurated in VBR");
        
        CalcVBR();

        break;
                
    case XING_VBR:
        DEBUGTRACE_MSG("time calculated in Xing header VBR");

        CalcXingVBR(xingFrames);
    }
}

void MP3Time::DetectFileType(FileType& fileType, unsigned long& xingFrames)
{
    InitMad();

    unsigned int count = 0;
    unsigned int prevBitrate = 0;
    bool bitrateChanged = false;

    for(;;) {

        if(mad_header_decode(&header_, &stream_) == -1) {
            if(MAD_RECOVERABLE(stream_.error))
                continue;
            else
                break;
        }

        if(count == 0) {

            prevBitrate = header_.bitrate;

            if(DetectXingHeader(xingFrames)) {

                fileType = XING_VBR;

                break;
            }
        }

        if(count < 114) {

            ++count;

            if(prevBitrate != header_.bitrate)
                bitrateChanged = true;

        } else if(count == 114 && !bitrateChanged) {

            fileType = CBR;

            break;
        }

        mad_timer_add(&duration_, header_.duration);
    }

    CleanMad();
}
