/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file tag_ogg.cc
/// @brief Class for process ogg tag header file.

#include "../include/tag_ogg.h"
#include "../include/exception.h"
#include "../include/trace.h"

/// @throw ErrnoException Exception on file open.
/// @throw OggException Exception on ov_*.
OGGTag::OGGTag(const std::string& filename)
{
    DEBUGTRACE("OGGTag::OGGTag");

    GetInfo(filename);

    if(!Find("TITLE", title_))
        title_ = Glib::path_get_basename(filename);

    totalTime_ = ConvertSecondToTimeString(totalTimeSeconds_);
}

void OGGTag::GetInfo(const std::string& filename)
{
    VORBIS::OVFile ovFile(filename.c_str());

    Insert(ovFile);

    totalTimeSeconds_ = static_cast<unsigned long int>(ovFile.GetTotalTime());
}

void OGGTag::Insert(VORBIS::OVFile& ovFile)
{
    DEBUGTRACE("OGGTag::Insert");

    if(ovFile.GetCommentsNum() > 0) {

        for(int i = 0; i < ovFile.GetCommentsNum(); ++i)
            InsertComment(ovFile.GetComments(i));
    }
}

void OGGTag::InsertComment(const std::string& comment)
{
    std::string::const_iterator equal(find_if(comment.begin(), comment.end(), IsCharEqual()));

    if(equal != comment.end()) {
        std::string key(comment.begin(), equal);
        // Recommendation is upper case, but it is just recommendation.
        // So convert all them to upper case.
        StringToUpper(key);
		
        tag_.insert(std::pair<std::string, std::string>(key, std::string(++equal, comment.end())));
    }
}

void OGGTag::ProcessYear()
{
    Find("DATE", year_);
}

void OGGTag::ProcessTrack()
{
    Find("TRACKNUMBER", track_);
}

void OGGTag::ProcessAlbum()
{
    Find("ALBUM", album_);
}

void OGGTag::ProcessGenre()
{
    Find("GENRE", genre_);
}

void OGGTag::ProcessArtist()
{
    Find("ARTIST", artist_);
}

void OGGTag::ProcessComment()
{
    Find("COMMENT", comment_);
}

/// @brief Find field in multimap and assign to output.
/// @param field Key name to find
/// @param output Output.
bool OGGTag::Find(const char* const field, Glib::ustring& output)
{
    typedef std::multimap<std::string, std::string>::iterator mapIter;
    typedef std::pair<mapIter, mapIter> mapIterPair;

    mapIterPair p(tag_.equal_range(field));

    if(p.first == p.second)
        return false;
	     
    for(; p.first != p.second; ++p.first) {
        output += p.first->second;
        output += ' ';
    }

    return true;
}

/// @brief Create OGGTag object.
///
/// Used in TagFactory.
std::auto_ptr<Tag> CreateOGGTag(const std::string& filename)
{
    return std::auto_ptr<Tag>(new OGGTag(filename));
}
