/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// wma decoder came from bmp-wma plugin.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file tag_wma.cc
/// @brief Class for process wma tag source file.

#include <string>

#include <glibmm.h>

#include "../include/tag_wma.h"
#include "../include/exception.h"
#include "../include/misc.h"
#include "../include/trace.h"

/// @throw WmaException Init failed.
WMATag::WMATag(const std::string& filename)
    : avCodec_(0),
      avCodecContext_(0),
      avFormatContext_(0)
{
    DEBUGTRACE("WMATag::WMATag");

    avcodec_init();
    avcodec_register_all();
    av_register_all();

    if(av_open_input_file(&avFormatContext_, filename.c_str(), 0, 0, 0) < 0)
        throw Exception::WmaException("av_open_input_file failed");
    
    for(int i = 0; i < avFormatContext_->nb_streams; ++i) {
        avCodecContext_ = &avFormatContext_->streams[i]->codec;
        if(avCodecContext_->codec_type == CODEC_TYPE_AUDIO)
            break;	
    }

    av_find_stream_info(avFormatContext_);

    try {

        AVCodecSelect();

    } catch(const Exception::SaguaException& ex) {

        av_close_input_file(avFormatContext_);

        throw;
    }

    GetTitle(filename);

    GetTime();
}

WMATag::~WMATag()
{
    avcodec_close(avCodecContext_);
    av_close_input_file(avFormatContext_);
}

void WMATag::ProcessYear()
{
    if(avFormatContext_->year != 0)
        IntegerToString(avFormatContext_->year, year_);
}

void WMATag::ProcessTrack()
{
    if(avFormatContext_->track != 0)
        IntegerToString(avFormatContext_->track, track_);
}

void WMATag::ProcessAlbum()
{
    if(avFormatContext_->album[0] != '\0')
        album_ = StringToUTF8::Instance().Convert(avFormatContext_->album);
}

void WMATag::ProcessGenre()
{
    if(avFormatContext_->genre[0] != '\0')
        genre_ = StringToUTF8::Instance().Convert(avFormatContext_->genre);
}

void WMATag::ProcessArtist()
{
    if(avFormatContext_->author[0] != '\0')
        artist_ = StringToUTF8::Instance().Convert(avFormatContext_->author);
}

void WMATag::ProcessComment()
{
    if(avFormatContext_->comment[0] != '\0')
        comment_ = StringToUTF8::Instance().Convert(avFormatContext_->comment);
}

/// @throw WmaException Exception on avcodec_find_decoder and open
void WMATag::AVCodecSelect()
{
    avCodec_ = avcodec_find_decoder(avCodecContext_->codec_id);
    if(!avCodec_)
        throw Exception::WmaException("avcodec_find_decoder failed");

    if(avcodec_open(avCodecContext_, avCodec_) < 0)
        throw Exception::WmaException("avcodec_open failed");
}

void WMATag::GetTime()
{
    if(avFormatContext_->duration != 0) {

        totalTimeSeconds_ = avFormatContext_->duration/1000000LL;
	    
        totalTime_ = ConvertSecondToTimeString(totalTimeSeconds_);
    }
}

void WMATag::GetTitle(const std::string& filename)
{
    if(avFormatContext_->title[0] != '\0')
        title_ = avFormatContext_->title;

    if(StringIsSpace(title_))
        title_ = StringToUTF8::Instance().Convert(Glib::path_get_basename(filename));
    else
        title_ = StringToUTF8::Instance().Convert(title_, Glib::path_get_basename(filename));
}

/// @brief Create WMATag object.
///
/// Used in TagFactory.
std::auto_ptr<Tag> CreateWMATag(const std::string& filename)
{
    return std::auto_ptr<Tag>(new WMATag(filename));
}
